"""FOFB devices."""

import time as _time
import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from ..namesys import SiriusPVName as _PVName
from ..search import BPMSearch as _BPMSearch, PSSearch as _PSSearch
from ..fofb.csdev import HLFOFBConst as _Const, NR_BPM

from .device import Device as _Device, ProptyDevice as _ProptyDevice, \
    Devices as _Devices
from .bpm import BPMLogicalTrigger
from .timing import Event
from .pwrsupply import PowerSupplyFC
from .psconv import StrengthConv


class _FOFBCtrlBase:
    """FOFB Ctrl base."""

    _devices = {
        f'SI{i:02d}': f'IA-{i:02d}RaBPM:BS-FOFBCtrl' for i in range(1, 21)}
    DEVICES = _get_namedtuple('DEVICES', *zip(*_devices.items()))


class FOFBCtrlRef(_Device, _FOFBCtrlBase):
    """FOFB reference orbit controller device."""

    _properties = (
        'RefOrb-SP', 'RefOrb-RB',
    )

    def __init__(self, devname):
        """Init."""
        # check if device exists
        if devname not in self.DEVICES:
            raise NotImplementedError(devname)

        # call base class constructor
        super().__init__(devname, properties=FOFBCtrlRef._properties)

    @property
    def ref(self):
        """Reference orbit, first half reference for X, second, for Y."""
        ref = self['RefOrb-RB']
        if ref is None:
            return None
        ref = ref.copy()
        # handle initial state of RefOrb PVs
        if len(ref) < 2*NR_BPM:
            value = _np.zeros(2*NR_BPM, dtype=int)
            value[:len(ref)] = ref
            ref = value
        return ref

    @ref.setter
    def ref(self, value):
        self['RefOrb-SP'] = _np.array(value, dtype=int)

    @property
    def refx(self):
        """Reference orbit X."""
        return self.ref[:NR_BPM]

    @refx.setter
    def refx(self, value):
        var = self.ref
        var[:NR_BPM] = _np.array(value, dtype=int)
        self.ref = var

    @property
    def refy(self):
        """Reference orbit Y."""
        return self.ref[NR_BPM:]

    @refy.setter
    def refy(self, value):
        var = self.ref
        var[NR_BPM:] = _np.array(value, dtype=int)
        self.ref = var

    def set_ref(self, value):
        """Set RefOrb."""
        self.ref = value
        return True

    def set_refx(self, value):
        """Set RefOrb X."""
        self.refx = value
        return True

    def set_refy(self, value):
        """Set RefOrb Y."""
        self.refy = value
        return True

    def check_ref(self, value):
        """Check whether RefOrb is equal to value."""
        if not _np.all(self.ref == value):
            return False
        return True

    def check_refx(self, value):
        """Check whether first half of RefOrb is equal to value."""
        return self._check_reforbit('x', value)

    def check_refy(self, value):
        """Check whether second half of RefOrb is equal to value."""
        return self._check_reforbit('y', value)

    def _check_reforbit(self, plane, value):
        refval = getattr(self, 'ref'+plane.lower())
        if not _np.all(refval == value):
            return False
        return True


class _DCCDevice(_ProptyDevice):
    """FOFB Diamond communication controller device."""

    DEF_TIMEOUT = 1
    DEF_FMC_BPMCNT = NR_BPM
    DEF_P2P_BPMCNT = 8

    _sync_sleep = 1.0  # [s]

    _properties = (
        'BPMId-SP', 'BPMId-RB', 'BPMCnt-Mon',
        'CCEnable-SP', 'CCEnable-RB',
        'TimeFrameLen-SP', 'TimeFrameLen-RB',
    )
    _properties_fmc = (
        'LinkPartnerCH0-Mon', 'LinkPartnerCH1-Mon',
        'LinkPartnerCH2-Mon', 'LinkPartnerCH3-Mon',
    )

    def __init__(self, devname, dccname):
        """Init."""
        self.dccname = dccname

        properties = _DCCDevice._properties
        if 'FMC' in self.dccname:
            properties += _DCCDevice._properties_fmc

        super().__init__(devname, dccname, properties=properties)
        prop2automon = [
            'BPMCnt-Mon', 'LinkPartnerCH0-Mon', 'LinkPartnerCH1-Mon',
            'LinkPartnerCH2-Mon', 'LinkPartnerCH3-Mon']
        for prop in prop2automon:
            self.set_auto_monitor(prop, True)

    @property
    def bpm_id(self):
        """BPM Id."""
        return self['BPMId-RB']

    @bpm_id.setter
    def bpm_id(self, value):
        self['BPMId-SP'] = value

    @property
    def bpm_count(self):
        """BPM count."""
        return self['BPMCnt-Mon']

    @property
    def cc_enable(self):
        """Communication enable."""
        return self['CCEnable-RB']

    @cc_enable.setter
    def cc_enable(self, value):
        self['CCEnable-SP'] = value

    @property
    def time_frame_len(self):
        """Time frame length."""
        return self['TimeFrameLen-RB']

    @time_frame_len.setter
    def time_frame_len(self, value):
        self['TimeFrameLen-SP'] = value

    @property
    def is_synced(self):
        """Is synchronized."""
        if not self.connected:
            return False
        cnt = self.DEF_FMC_BPMCNT if 'FMC' in self.dccname \
            else self.DEF_P2P_BPMCNT
        return self['BPMCnt-Mon'] == cnt


class FOFBCtrlDCC(_DCCDevice, _FOFBCtrlBase):
    """FOFBCtrl DCC device."""

    class PROPDEVICES:
        """DCC devices."""

        FMC = 'DCCFMC'
        P2P = 'DCCP2P'
        ALL = (FMC, P2P)

    def __init__(self, devname, dccname):
        """Init."""
        if devname not in self.DEVICES:
            raise NotImplementedError(devname)
        if dccname not in self.PROPDEVICES.ALL:
            raise NotImplementedError(dccname)
        super().__init__(devname, dccname)

    @property
    def linkpartners(self):
        """Return linked partners."""
        linkpart_props = [
            'LinkPartnerCH0-Mon', 'LinkPartnerCH1-Mon',
            'LinkPartnerCH2-Mon', 'LinkPartnerCH3-Mon']
        return set(self[prop] for prop in linkpart_props)


class BPMDCC(_DCCDevice):
    """BPM DCC device."""

    def __init__(self, devname):
        """Init."""
        if not _BPMSearch.is_valid_devname(devname):
            raise NotImplementedError(devname)
        super().__init__(devname, 'DCCP2P')


class FamFOFBControllers(_Devices):
    """Family of FOFBCtrl and related BPM devices."""

    DEF_TIMEOUT = 10  # [s]
    DEF_DCC_TIMEFRAMELEN = 5000
    DEF_BPMTRIG_RCVSRC = 0
    DEF_BPMTRIG_RCVIN = 5
    BPM_TRIGS_IDS = [1, 2, 20]
    FOFBCTRL_BPMID_OFFSET = 480

    def __init__(self):
        """Init."""
        # FOFBCtrl Refs and DCCs
        bpmids = _np.array(
            [self.FOFBCTRL_BPMID_OFFSET - 1 + i for i in range(1, 21)])
        lpcw = _np.roll(bpmids, 1)
        lpaw = _np.roll(bpmids, -1)
        self._ctl_ids, self._ctl_part = dict(), dict()
        self._ctl_refs, self._ctl_dccs = dict(), dict()
        for idx, ctl in enumerate(_FOFBCtrlBase.DEVICES):
            self._ctl_ids[ctl] = bpmids[idx]
            self._ctl_part[ctl] = {lpcw[idx], lpaw[idx]}
            self._ctl_refs[ctl] = FOFBCtrlRef(ctl)
            for dcc in FOFBCtrlDCC.PROPDEVICES.ALL:
                self._ctl_dccs[ctl + ':' + dcc] = FOFBCtrlDCC(ctl, dcc)
        # BPM DCCs and triggers
        self._bpmnames = _BPMSearch.get_names({'sec': 'SI', 'dev': 'BPM'})
        bpmids = [((i + 1) // 2) * 2 % 160 for i in range(NR_BPM)]
        self._bpm_dccs, self._bpm_trgs, self._bpm_ids = dict(), dict(), dict()
        for idx, bpm in enumerate(self._bpmnames):
            self._bpm_ids[bpm] = bpmids[idx]
            self._bpm_dccs[bpm] = BPMDCC(bpm)
            for trig in self.BPM_TRIGS_IDS:
                trigname = bpm + ':TRIGGER' + str(trig)
                self._bpm_trgs[trigname] = BPMLogicalTrigger(bpm, trig)
        # fofb event
        self._evt_fofb = Event('FOFBS')

        devices = list()
        devices.extend(self._ctl_refs.values())
        devices.extend(self._ctl_dccs.values())
        devices.extend(self._bpm_dccs.values())
        devices.extend(self._bpm_trgs.values())
        devices.append(self._evt_fofb)

        super().__init__('SI-Glob:BS-FOFB', devices)

    def set_reforb(self, value):
        """Set RefOrb for all FOFB controllers."""
        for ctrl in self._ctl_refs.values():
            ctrl.set_ref(value)
        return True

    def set_reforbx(self, value):
        """Set RefOrbX for all FOFB controllers."""
        return self._set_reforb('x', value)

    def set_reforby(self, value):
        """Set RefOrbY for all FOFB controllers."""
        return self._set_reforb('y', value)

    def _set_reforb(self, plane, value):
        for ctrl in self._ctl_refs.values():
            fun = getattr(ctrl, 'set_ref' + plane.lower())
            fun(value)
        return True

    def check_reforb(self, value):
        """Check whether RefOrb is equal to value."""
        if not self.connected:
            return False
        for ctrl in self._ctl_refs.values():
            if not ctrl.check_ref(value):
                return False
        return True

    def check_reforbx(self, value):
        """Check whether RefOrbX is equal to value."""
        return self._check_reforb('x', value)

    def check_reforby(self, value):
        """Check whether RefOrbY is equal to value."""
        return self._check_reforb('y', value)

    def _check_reforb(self, plane, value):
        if not self.connected:
            return False
        for ctrl in self._ctl_refs.values():
            fun = getattr(ctrl, 'check_ref' + plane.lower())
            if not fun(value):
                return False
        return True

    @property
    def bpm_id(self):
        """Return DCC BPMIds."""
        if not self.connected:
            return False
        bpmids = dict()
        for dev in self._ctl_dccs.values():
            bpmids[dev.pv_object('BPMId-RB').pvname] = dev.bpm_id
        for dev in self._bpm_dccs.values():
            bpmids[dev.pv_object('BPMId-RB').pvname] = dev.bpm_id
        return bpmids

    @property
    def bpm_id_configured(self):
        """Check whether DCC BPMIds are configured."""
        if not self.connected:
            return False
        for dcc, dev in self._ctl_dccs.items():
            ctl = _PVName(dcc).device_name
            if not dev.bpm_id == self._ctl_ids[ctl]:
                return False
        for bpm, dev in self._bpm_dccs.items():
            if not dev.bpm_id == self._bpm_ids[bpm]:
                return False
        return True

    def cmd_config_bpm_id(self):
        """Command to configure DCC BPMIds."""
        if not self.connected:
            return False
        for dcc, dev in self._ctl_dccs.items():
            ctl = _PVName(dcc).device_name
            dev.bpm_id = self._ctl_ids[ctl]
        for bpm, dev in self._bpm_dccs.items():
            dev.bpm_id = self._bpm_ids[bpm]
        return True

    @property
    def linkpartners(self):
        """Return link partners."""
        if not self.connected:
            return False
        partners = dict()
        for dev in self._ctl_dccs.values():
            if 'FMC' not in dev.dccname:
                continue
            partners[dev.devname] = dev.linkpartners
        return partners

    @property
    def linkpartners_connected(self):
        """Check whether adjacent partners are connected."""
        if not self.connected:
            return False
        for dcc, dev in self._ctl_dccs.items():
            if 'FMC' not in dev.dccname:
                continue
            ctl = _PVName(dcc).device_name
            nrpart = len(self._ctl_part[ctl])
            if not len(dev.linkpartners & self._ctl_part[ctl]) == nrpart:
                return False
        return True

    @property
    def bpm_count(self):
        """Return DCC BPMCnt."""
        if not self.connected:
            return False
        bpmids = dict()
        for dev in self._ctl_dccs.values():
            bpmids[dev.pv_object('BPMCnt-Mon').pvname] = dev.bpm_count
        for dev in self._bpm_dccs.values():
            bpmids[dev.pv_object('BPMCnt-Mon').pvname] = dev.bpm_count
        return bpmids

    def cmd_sync_net(self, bpms=None, timeout=DEF_TIMEOUT):
        """Command to synchronize DCCs."""
        alldccs = list(self._ctl_dccs.values()) + list(self._bpm_dccs.values())
        enbdccs = list(self._ctl_dccs.values())
        if bpms is None:
            bpms = self._bpmnames
        for bpm in bpms:
            enbdccs.append(self._bpm_dccs[bpm])

        self._set_devices_propty(alldccs, 'CCEnable-SP', 0)
        if not self._wait_devices_propty(
                alldccs, 'CCEnable-RB', 0, timeout=timeout/2):
            return False
        self._set_devices_propty(enbdccs, 'CCEnable-SP', 1)
        if not self._wait_devices_propty(
                enbdccs, 'CCEnable-RB', 1, timeout=timeout/2):
            return False
        self._evt_fofb.cmd_external_trigger()
        return True

    def check_net_synced(self, bpms=None):
        """Check whether DCCs are synchronized."""
        if not self.connected:
            return False
        if bpms is None:
            bpms = self._bpmnames
        for dev in self._ctl_dccs.values():
            if dev.dccname != 'DCCFMC':
                continue
            if not dev.bpm_count == len(bpms):
                return False
        for bpm in bpms:
            if not self._bpm_dccs[bpm].cc_enable == 1:
                return False
        return True

    @property
    def time_frame_len(self):
        """Time frame length.

        Returns:
            length (numpy.ndarray, 20*2+160):
                time frame length for each FOFB controller DCC and for
                each BPM DCC.
        """
        dccs = list(self._ctl_dccs.values()) + list(self._bpm_dccs.values())
        return _np.array([d.time_frame_len for d in dccs])

    def set_time_frame_len(
            self, value=DEF_DCC_TIMEFRAMELEN, timeout=DEF_TIMEOUT):
        """Set DCCs TimeFrameLen."""
        dccs = list(self._ctl_dccs.values()) + list(self._bpm_dccs.values())
        self._set_devices_propty(dccs, 'TimeFrameLen-SP', value)
        if not self._wait_devices_propty(
                dccs, 'TimeFrameLen-RB', value, timeout=timeout):
            return False
        return True

    @property
    def bpm_trigs_configuration(self):
        """Return BPM logical triggers configuration."""
        if not self.connected:
            return False
        conf = dict()
        for dev in self._bpm_trgs.values():
            conf[dev.pv_object('RcvSrc-Sts').pvname] = dev.receiver_source
            conf[dev.pv_object('RcvInSel-RB').pvname] = dev.receiver_in_sel
        return conf

    @property
    def bpm_trigs_configured(self):
        """Check whether all BPM triggers are configured."""
        if not self.connected:
            return False
        for dev in self._bpm_trgs.values():
            if not dev.receiver_source == self.DEF_BPMTRIG_RCVSRC:
                return False
            if not dev.receiver_in_sel == self.DEF_BPMTRIG_RCVIN:
                return False
        return True

    def cmd_config_bpm_trigs(self, timeout=DEF_TIMEOUT):
        """Command to configure BPM triggers."""
        devs = list(self._bpm_trgs.values())
        rsrc = self.DEF_BPMTRIG_RCVSRC
        rins = self.DEF_BPMTRIG_RCVIN
        self._set_devices_propty(devs, 'RcvSrc-Sel', rsrc)
        if not self._wait_devices_propty(
                devs, 'RcvSrc-Sts', rsrc, timeout=timeout/2):
            return False
        self._set_devices_propty(devs, 'RcvInSel-SP', rins)
        if not self._wait_devices_propty(
                devs, 'RcvInSel-RB', rins, timeout=timeout/2):
            return False
        return True


class FamFastCorrs(_Devices):
    """Family of FOFB fast correctors."""

    DEF_TIMEOUT = 10  # [s]
    OPMODE_SEL = PowerSupplyFC.OPMODE_SEL
    OPMODE_STS = PowerSupplyFC.OPMODE_STS
    DEF_ATOL_INVRESPMATROW = 2**-17
    DEF_ATOL_FOFBACCGAIN = 2**-12
    DEF_ATOL_FOFBACCSAT = 2e-2
    DEF_ATOL_CURRENT_RB = 1e-6
    DEF_ATOL_CURRENT_MON = 2e-2

    def __init__(self, psnames=None):
        """Init."""
        if not psnames:
            chn = _PSSearch.get_psnames({'sec': 'SI', 'dev': 'FCH'})
            cvn = _PSSearch.get_psnames({'sec': 'SI', 'dev': 'FCV'})
            psnames = chn + cvn
        self._psnames = psnames
        self._psdevs = [PowerSupplyFC(psn) for psn in self._psnames]
        self._psconv = [StrengthConv(psn, 'Ref-Mon', auto_mon=True)
                        for psn in self._psnames]
        super().__init__('SI-Glob:PS-FCHV', self._psdevs + self._psconv)

    @property
    def psnames(self):
        """PS name list."""
        return list(self._psnames)

    @property
    def psdevs(self):
        """PS device list."""
        return self._psdevs

    @property
    def psconvs(self):
        """PS conversion device list."""
        return self._psconv

    @property
    def pwrstate(self):
        """Power State.

        Returns:
            state (numpy.ndarray, 160):
                PwrState for each power supply.
        """
        return _np.array([p.pwrstate for p in self._psdevs])

    @property
    def opmode(self):
        """Operation Mode.

        Returns:
            mode (numpy.ndarray, 160):
                OpMode for each power supply.
        """
        return _np.array([p.opmode for p in self._psdevs])

    @property
    def current(self):
        """Current readback.

        Returns:
            current (numpy.ndarray, 160):
                OpMode for each power supply.
        """
        return _np.array([p.current for p in self._psdevs])

    @property
    def current_mon(self):
        """Implemented current.

        Returns:
            current (numpy.ndarray, 160):
                OpMode for each power supply.
        """
        return _np.array([p.current_mon for p in self._psdevs])

    @property
    def fofbacc_gain(self):
        """FOFB pre-accumulator gain.

        Returns:
            gain (numpy.ndarray, 160):
                FOFB pre-accumulator gain for each power supply.
        """
        return _np.array([p.fofbacc_gain for p in self._psdevs])

    @property
    def fofbacc_freeze(self):
        """FOFB pre-accumulator freeze state.

        Returns:
            gain (numpy.ndarray, 160):
                FOFB pre-accumulator freeze state for each power supply.
        """
        return _np.array([p.fofbacc_freeze for p in self._psdevs])

    @property
    def fofbacc_satmax(self):
        """FOFB pre-accumulator maximum saturation current [A].

        Returns:
            gain (numpy.ndarray, 160):
                FOFB pre-accumulator maximum saturation current
                for each power supply.
        """
        return _np.array([p.fofbacc_satmax for p in self._psdevs])

    @property
    def fofbacc_satmin(self):
        """FOFB pre-accumulator minimum saturation current [A].

        Returns:
            gain (numpy.ndarray, 160):
                FOFB pre-accumulator minimum saturation current
                for each power supply.
        """
        return _np.array([p.fofbacc_satmin for p in self._psdevs])

    @property
    def curr_gain(self):
        """Current gain.

        Returns:
            gain (numpy.ndarray, 160):
                current gain for each power supply.
        """
        return _np.array([p.curr_gain for p in self._psdevs])

    @property
    def strength_2_current_factor(self):
        """Strength to current convertion factor.

        Returns:
            factor (numpy.ndarray, NR_BPM):
                convertion factor for each power supply.
        """
        return _np.array(
            [p.conv_strength_2_current(1, strengths_dipole=3.0)
             for p in self._psconv])

    def set_pwrstate(
            self, state, psnames=None, psindices=None):
        """Command to set power supply pwrstate."""
        devs = self._get_devices(psnames, psindices)
        self._set_devices_propty(devs, 'PwrState-Sel', state)
        return True

    def check_pwrstate(
            self, state, psnames=None, psindices=None,
            timeout=DEF_TIMEOUT):
        """Check whether power supplies are in desired pwrstate."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        return self._wait_devices_propty(
            devs, 'PwrState-Sts', state, timeout=timeout)

    def set_opmode(
            self, opmode, psnames=None, psindices=None):
        """Command to set power supply opmode."""
        devs = self._get_devices(psnames, psindices)
        self._set_devices_propty(devs, 'OpMode-Sel', opmode)
        return True

    def check_opmode(
            self, opmode, psnames=None, psindices=None,
            timeout=DEF_TIMEOUT):
        """Check whether power supplies are in desired opmode."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        return self._wait_devices_propty(
            devs, 'OpMode-Sts', opmode, timeout=timeout)

    def set_current(self, values, psnames=None, psindices=None):
        """Set power supply current."""
        devs = self._get_devices(psnames, psindices)
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        for i, dev in enumerate(devs):
            dev.current = values[i]
        return True

    def check_current(
            self, values, psnames=None, psindices=None,
            atol=DEF_ATOL_CURRENT_RB):
        """Check whether power supplies have desired current."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        impltd = _np.asarray([d.current for d in devs])
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        if _np.allclose(values, impltd, atol=atol):
            return True
        return False

    def check_current_mon(
            self, values, psnames=None, psindices=None,
            atol=DEF_ATOL_CURRENT_MON):
        """Check whether power supplies have desired implemented current."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        impltd = _np.asarray([d.current_mon for d in devs])
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        if _np.allclose(values, impltd, atol=atol):
            return True
        return False

    def set_invrespmat_row(self, values, psnames=None, psindices=None):
        """Command to set power supply correction coefficients value."""
        if not isinstance(values, (list, tuple, _np.ndarray)):
            raise ValueError('Value must be iterable.')
        devs = self._get_devices(psnames, psindices)
        if any([len(v) != 2*NR_BPM for v in values]):
            raise ValueError(f'Value must have size {2*NR_BPM}.')
        for i, dev in enumerate(devs):
            dev.invrespmat_row = values[i]
        return True

    def check_invrespmat_row(
            self, values, psnames=None, psindices=None,
            atol=DEF_ATOL_INVRESPMATROW):
        """Check power supplies correction coefficients."""
        if not self.connected:
            return False
        if not isinstance(values, (list, tuple, _np.ndarray)):
            raise ValueError('Value must be iterable.')
        values = _np.asarray(values)
        devs = self._get_devices(psnames, psindices)
        if not values.shape[0] == len(devs):
            raise ValueError('Values and indices must have the same size.')
        impltd = _np.asarray([d.invrespmat_row for d in devs])
        if _np.allclose(values, impltd, atol=atol):
            return True
        return False

    def set_fofbacc_gain(self, values, psnames=None, psindices=None):
        """Command to set power supply correction gain."""
        if not isinstance(values, (list, tuple, _np.ndarray)):
            raise ValueError('Value must be iterable.')
        devs = self._get_devices(psnames, psindices)
        for i, dev in enumerate(devs):
            dev.fofbacc_gain = values[i]
        return True

    def check_fofbacc_gain(
            self, values, psnames=None, psindices=None,
            atol=DEF_ATOL_FOFBACCGAIN):
        """Check whether power supplies have desired correction gain."""
        if not self.connected:
            return False
        if not isinstance(values, (list, tuple, _np.ndarray)):
            raise ValueError('Value must be iterable.')
        values = _np.asarray(values)
        devs = self._get_devices(psnames, psindices)
        impltd = _np.asarray([d.fofbacc_gain for d in devs])
        if _np.allclose(values, impltd, atol=atol):
            return True
        return False

    def set_fofbacc_freeze(self, values, psnames=None, psindices=None):
        """Command to set power supply pre-accumulator freeze state."""
        devs = self._get_devices(psnames, psindices)
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        for i, dev in enumerate(devs):
            dev.fofbacc_freeze = values[i]
        return True

    def check_fofbacc_freeze(
            self, values, psnames=None, psindices=None,
            timeout=DEF_TIMEOUT):
        """Check whether power supplies have desired freeze state."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        values = list(values)
        return self._wait_devices_propty(
            devs, 'FOFBAccFreeze-Sts', values, timeout=timeout)

    def set_fofbacc_satmax(self, values, psnames=None, psindices=None):
        """Set power supply pre-accumulator max.saturation current."""
        devs = self._get_devices(psnames, psindices)
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        for i, dev in enumerate(devs):
            dev.fofbacc_satmax = values[i]
        return True

    def check_fofbacc_satmax(
            self, values, psnames=None, psindices=None,
            atol=DEF_ATOL_FOFBACCSAT):
        """Check whether power supplies have desired max.saturation value."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        impltd = _np.asarray([d.fofbacc_satmax for d in devs])
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        if _np.allclose(values, impltd, atol=atol):
            return True
        return False

    def set_fofbacc_satmin(self, values, psnames=None, psindices=None):
        """Set power supply pre-accumulator min.saturation current."""
        devs = self._get_devices(psnames, psindices)
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        for i, dev in enumerate(devs):
            dev.fofbacc_satmin = values[i]
        return True

    def check_fofbacc_satmin(
            self, values, psnames=None, psindices=None,
            atol=DEF_ATOL_FOFBACCSAT):
        """Check whether power supplies have desired min.saturation value."""
        if not self.connected:
            return False
        devs = self._get_devices(psnames, psindices)
        impltd = _np.asarray([d.fofbacc_satmin for d in devs])
        if isinstance(values, (int, float, bool)):
            values = len(devs) * [values]
        if _np.allclose(values, impltd, atol=atol):
            return True
        return False

    def cmd_fofbacc_clear(self, psnames=None, psindices=None):
        """Send clear power supplies pre-accumulator."""
        for dev in self._get_devices(psnames, psindices):
            dev.cmd_fofbacc_clear()
        return True

    # ----- private methods -----

    def _get_devices(self, names, indices):
        if names:
            indices = [self._psnames.index(psn) for psn in names]
        elif indices is not None and \
                not isinstance(indices, (list, tuple, _np.ndarray)):
            indices = [indices, ]
        if indices is None:
            indices = [i for i in range(len(self._psnames))]
        return [self._psdevs[i] for i in indices]


class HLFOFB(_Device):
    """Control high level FOFB IOC."""

    class DEVICES:
        """Devices names."""

        SI = 'SI-Glob:AP-FOFB'
        ALL = (SI, )

    _properties = (
        'LoopState-Sel', 'LoopState-Sts',
        'LoopGainH-SP', 'LoopGainH-RB', 'LoopGainH-Mon',
        'LoopGainV-SP', 'LoopGainV-RB', 'LoopGainV-Mon',
        'CorrStatus-Mon', 'CorrConfig-Cmd',
        'CorrSetPwrStateOn-Cmd', 'CorrSetOpModeManual-Cmd',
        'CorrSetAccFreezeDsbl-Cmd', 'CorrSetAccFreezeEnbl-Cmd',
        'CorrSetAccClear-Cmd', 'CorrSetCurrZero-Cmd',
        'CHAccSatMax-SP', 'CHAccSatMax-RB',
        'CVAccSatMax-SP', 'CVAccSatMax-RB',
        'FOFBCtrlStatus-Mon', 'FOFBCtrlConfBPMId-Cmd',
        'FOFBCtrlSyncNet-Cmd', 'FOFBCtrlSyncRefOrb-Cmd',
        'FOFBCtrlConfTFrameLen-Cmd', 'FOFBCtrlConfBPMLogTrg-Cmd',
        'KickBufferSize-SP', 'KickBufferSize-RB', 'KickBufferSize-Mon',
        'KickCH-Mon', 'KickCV-Mon',
        'RefOrbX-SP', 'RefOrbX-RB', 'RefOrbY-SP', 'RefOrbY-RB',
        'RefOrbHwX-Mon', 'RefOrbHwY-Mon',
        'BPMXEnblList-SP', 'BPMXEnblList-RB',
        'BPMYEnblList-SP', 'BPMYEnblList-RB',
        'CHEnblList-SP', 'CHEnblList-RB',
        'CVEnblList-SP', 'CVEnblList-RB',
        'UseRF-Sel', 'UseRF-Sts',
        'MinSingValue-SP', 'MinSingValue-RB',
        'TikhonovRegConst-SP', 'TikhonovRegConst-RB',
        'SingValuesRaw-Mon', 'SingValues-Mon', 'NrSingValues-Mon',
        'RespMat-SP', 'RespMat-RB', 'RespMat-Mon', 'InvRespMat-Mon',
        'SingValuesHw-Mon', 'RespMatHw-Mon', 'InvRespMatHw-Mon',
        'InvRespMatNormMode-Sel', 'InvRespMatNormMode-Sts',
        'CorrCoeffs-Mon', 'CorrGains-Mon',
        'MeasRespMat-Cmd', 'MeasRespMat-Mon',
        'MeasRespMatKickCH-SP', 'MeasRespMatKickCH-RB',
        'MeasRespMatKickCV-SP', 'MeasRespMatKickCV-RB',
        'MeasRespMatKickRF-SP', 'MeasRespMatKickRF-RB',
        'MeasRespMatWait-SP', 'MeasRespMatWait-RB',
    )

    _default_timeout_respm = 2 * 60 * 60  # [s]

    def __init__(self, devname=None):
        """Init."""
        # check if device exists
        if devname is None:
            devname = HLFOFB.DEVICES.SI
        if devname not in HLFOFB.DEVICES.ALL:
            raise NotImplementedError(devname)

        self._data = _Const()

        # call base class constructor
        super().__init__(devname, properties=self._properties)

    @property
    def data(self):
        """IOC constants."""
        return self._data

    @property
    def loop_state(self):
        """Loop state."""
        return self['LoopState-Sts']

    @loop_state.setter
    def loop_state(self, value):
        self['LoopState-Sel'] = value

    @property
    def loop_gain_h(self):
        """Loop gain H."""
        return self['LoopGainH-RB']

    @loop_gain_h.setter
    def loop_gain_h(self, value):
        self['LoopGainH-SP'] = value

    @property
    def loop_gain_h_mon(self):
        """Implemented horizontal loop gain."""
        return self['LoopGainH-Mon']

    @property
    def loop_gain_v(self):
        """Loop gain V."""
        return self['LoopGainV-RB']

    @loop_gain_v.setter
    def loop_gain_v(self, value):
        self['LoopGainV-SP'] = value

    @property
    def loop_gain_v_mon(self):
        """Implemented vertical loop gain."""
        return self['LoopGainV-Mon']

    @property
    def corr_status(self):
        """Corrector status."""
        return self['CorrStatus-Mon']

    def cmd_corr_config(self):
        """Command to configure correctors in use."""
        self['CorrConfig-Cmd'] = 1
        return True

    def cmd_corr_set_pwrstate_on(self):
        """Command to set all corrector pwrstate to on."""
        self['CorrSetCurrZero-Cmd'] = 1
        return True

    def cmd_corr_set_opmode_manual(self):
        """Command to set all corrector opmode to manual."""
        self['CorrSetOpModeManual-Cmd'] = 1
        return True

    def cmd_corr_set_accfreeze_dsbl(self):
        """Command to set all corrector FOFBAccFreeze to Dsbl."""
        self['CorrSetAccFreezeDsbl-Cmd'] = 1
        return True

    def cmd_corr_set_accfreeze_enbl(self):
        """Command to set all corrector FOFBAccFreeze to Enbl."""
        self['CorrSetAccFreezeEnbl-Cmd'] = 1
        return True

    def cmd_corr_accclear(self):
        """Command to clear all corrector accumulator."""
        self['CorrSetAccClear-Cmd'] = 1
        return True

    def cmd_corr_set_current_zero(self):
        """Command to set correctors current to zero."""
        self['CorrSetCurrZero-Cmd'] = 1
        return True

    @property
    def ch_accsatmax(self):
        """CH accumulator maximum saturation limit."""
        return self['CHAccSatMax-RB']

    @ch_accsatmax.setter
    def ch_accsatmax(self, value):
        self['CHAccSatMax-SP'] = value

    @property
    def cv_accsatmax(self):
        """CH accumulator maximum saturation limit."""
        return self['CVAccSatMax-RB']

    @cv_accsatmax.setter
    def cv_accsatmax(self, value):
        self['CVAccSatMax-SP'] = value

    @property
    def fofbctrl_status(self):
        """FOFB controller status."""
        return self['FOFBCtrlStatus-Mon']

    def cmd_fofbctrl_conf_bpmid(self):
        """Command to configure all FOFB DCC BPMIds."""
        self['FOFBCtrlConfBPMId-Cmd'] = 1
        return True

    def cmd_fofbctrl_syncnet(self):
        """Command to sync FOFB controller net."""
        self['FOFBCtrlSyncNet-Cmd'] = 1
        return True

    def cmd_fofbctrl_syncreforb(self):
        """Command to sync FOFB controller RefOrb."""
        self['FOFBCtrlSyncRefOrb-Cmd'] = 1
        return True

    def cmd_fofbctrl_conf_timeframelen(self):
        """Command to configure all FOFB controller TimeFrameLen."""
        self['FOFBCtrlConfTFrameLen-Cmd'] = 1
        return True

    def cmd_fofbctrl_conf_bpmlogtrig(self):
        """Command to configure all BPM logical triggers related to FOFB."""
        self['FOFBCtrlConfBPMLogTrg-Cmd'] = 1
        return True

    @property
    def kick_buffer_size_mon(self):
        """Return actual kicks buffer size."""
        return self['KickBufferSize-Mon']

    @property
    def kick_buffer_size(self):
        """Return kicks buffer size."""
        return self['KickBufferSize-RB']

    @kick_buffer_size.setter
    def kick_buffer_size(self, value):
        self['KickBufferSize-SP'] = max(1, int(value))

    @property
    def kickch(self):
        """Return average of CH kicks."""
        return self['KickCH-Mon']

    @property
    def kickcv(self):
        """Return average of CV kicks."""
        return self['KickCV-Mon']

    @property
    def refx(self):
        """RefOrb X."""
        return self['RefOrbX-RB']

    @refx.setter
    def refx(self, value):
        self['RefOrbX-SP'] = value

    @property
    def refx_hw(self):
        """RefOrb X in hardware units."""
        return self['RefOrbHwX-Mon']

    @property
    def refy(self):
        """RefOrb Y."""
        return self['RefOrbY-RB']

    @refy.setter
    def refy(self, value):
        self['RefOrbY-SP'] = value

    @property
    def refy_hw(self):
        """RefOrb Y in hardware units."""
        return self['RefOrbHwY-Mon']

    @property
    def bpmxenbl(self):
        """BPM X enable list."""
        return self['BPMXEnblList-RB']

    @bpmxenbl.setter
    def bpmxenbl(self, value):
        self['BPMXEnblList-SP'] = value

    @property
    def bpmyenbl(self):
        """BPM Y enable list."""
        return self['BPMYEnblList-RB']

    @bpmyenbl.setter
    def bpmyenbl(self, value):
        self['BPMYEnblList-SP'] = value

    @property
    def chenbl(self):
        """CH enable list."""
        return self['CHEnblList-RB']

    @chenbl.setter
    def chenbl(self, value):
        self['CHEnblList-SP'] = value

    @property
    def cvenbl(self):
        """CV enable list."""
        return self['CVEnblList-RB']

    @cvenbl.setter
    def cvenbl(self, value):
        self['CVEnblList-SP'] = value

    @property
    def rfenbl(self):
        """Use RF in RespMat calculation."""
        return self['UseRF-Sts']

    @rfenbl.setter
    def rfenbl(self, value):
        self['UseRF-Sel'] = value

    @property
    def singval_min(self):
        """Minimum singular value."""
        return self['MinSingValue-RB']

    @singval_min.setter
    def singval_min(self, value):
        self['MinSingValue-SP'] = value

    @property
    def tikhonov_reg_const(self):
        """Tikhonov regularization constant."""
        return self['TikhonovRegConst-RB']

    @tikhonov_reg_const.setter
    def tikhonov_reg_const(self, value):
        self['TikhonovRegConst-SP'] = value

    @property
    def singvalsraw_mon(self):
        """Raw singular values of physical unit respmat."""
        return self['SingValuesRaw-Mon']

    @property
    def singvals_mon(self):
        """Singular values of physical unit respmat in use."""
        return self['SingValues-Mon']

    @property
    def nr_singvals(self):
        """Number of singular values."""
        return self['NrSingValues-Mon']

    @property
    def respmat(self):
        """RespMat in physical units."""
        return self['RespMat-RB'].reshape(self._data.nr_bpms*2, -1)

    @respmat.setter
    def respmat(self, mat):
        self['RespMat-SP'] = _np.array(mat).ravel()

    @property
    def respmat_mon(self):
        """RespMat in physical units in use."""
        return self['RespMat-Mon'].reshape(self._data.nr_bpms*2, -1)

    @property
    def invrespmat_mon(self):
        """InvRespMat in physical units in use."""
        return self['InvRespMat-Mon'].reshape(-1, self._data.nr_bpms*2)

    @property
    def singvalshw_mon(self):
        """Singular values of hardware unit respmat."""
        return self['SingValuesHw-Mon']

    @property
    def respmathw_mon(self):
        """RespMat in hardware units in use."""
        return self['RespMatHw-Mon'].reshape(self._data.nr_bpms*2, -1)

    @property
    def invrespmathw_mon(self):
        """InvRespMat in physical units in use."""
        return self['InvRespMatHw-Mon'].reshape(-1, self._data.nr_bpms*2)

    @property
    def invrespmat_normmode(self):
        """InvRespMat normalization mode."""
        return self['InvRespMatNormMode-Sts']

    @invrespmat_normmode.setter
    def invrespmat_normmode(self, value):
        self['InvRespMatNormMode-Sel'] = value

    @property
    def corrcoeffs(self):
        """InvRespMatRow setpoint for all correctors."""
        return self['CorrCoeffs-Mon'].reshape(-1, self._data.nr_bpms*2)

    @property
    def corrgains(self):
        """FOFBAccGain setpoint for all correctors."""
        return self['CorrGains-Mon']

    def cmd_measrespmat_start(self):
        """Command to start response matrix measure."""
        self['MeasRespMat-Cmd'] = self._data.MeasRespMatCmd.Start
        return True

    def cmd_measrespmat_stop(self):
        """Command to stop response matrix measure."""
        self['MeasRespMat-Cmd'] = self._data.MeasRespMatCmd.Stop
        return True

    def cmd_measrespmat_reset(self):
        """Command to reset response matrix measure."""
        self['MeasRespMat-Cmd'] = self._data.MeasRespMatCmd.Reset
        return True

    @property
    def measrespmat_mon(self):
        """RespMat measure status."""
        return self['MeasRespMat-Mon']

    @property
    def measrespmat_kickch(self):
        """RespMat measure CH kick."""
        return self['MeasRespMatKickCH-RB']

    @measrespmat_kickch.setter
    def measrespmat_kickch(self, value):
        self['MeasRespMatKickCH-SP'] = value

    @property
    def measrespmat_kickcv(self):
        """RespMat measure CV kick."""
        return self['MeasRespMatKickCV-RB']

    @measrespmat_kickcv.setter
    def measrespmat_kickcv(self, value):
        self['MeasRespMatKickCV-SP'] = value

    @property
    def measrespmat_kickrf(self):
        """RespMat measure RF kick."""
        return self['MeasRespMatKickRF-RB']

    @measrespmat_kickrf.setter
    def measrespmat_kickrf(self, value):
        self['MeasRespMatKickRF-SP'] = value

    @property
    def measrespmat_wait(self):
        """RespMat measure wait interval."""
        return self['MeasRespMatWait-RB']

    @measrespmat_wait.setter
    def measrespmat_wait(self, value):
        self['MeasRespMatWait-SP'] = value

    def wait_respm_meas(self, timeout=None):
        """Wait for response matrix measure."""
        timeout = timeout or self._default_timeout_respm
        return self._wait(
            'MeasRespMat-Mon', self._data.MeasRespMatMon.Measuring,
            timeout=timeout, comp='ne')



#!/usr/bin/env python-sirius

"""Unittest module for search.py."""
from unittest import mock

import os
from siriuspy.search import ps_search
from siriuspy.search import ma_search
from siriuspy.search import hl_time_search
from siriuspy.search import ll_time_search
from siriuspy.search import bpms_search
from siriuspy.magnet import excdata


_path = os.path.abspath(os.path.dirname(__file__))


# --- auxiliar methods ---
def _read_test_file(path):
    """Read a file."""
    prefix = _path + '/test_data/servweb/'
    with open(prefix + path, "r") as fd:
        return fd.read()


def _ps_pstype_data_read(path):
    """Read a file."""
    prefix = _path + '/test_data/servweb/pwrsupply/pstypes-data/'
    with open(prefix + path, "r") as fd:
        return fd.read()


def _magnets_excitation_data_read(label):
    """Read a file."""
    prefix = _path + '/test_data/servweb/magnet/excitation-data/'
    with open(prefix + label, "r") as fd:
        return fd.read()


# --- Define Mocks ---

# --- mock objects ps_search _web ---
_p = mock.patch.object(ps_search, '_web', autospec=True)
mock1 = _p.start()
mock1.ps_pstypes_names_read.return_value = _read_test_file(
    'pwrsupply/pstypes-names.txt')
mock1.ps_pstype_data_read.side_effect = _ps_pstype_data_read
mock1.ps_pstype_setpoint_limits.return_value = _read_test_file(
    'pwrsupply/pstypes-setpoint-limits.txt')
mock1.ps_siggen_configuration_read.return_value = _read_test_file(
    'pwrsupply/siggen-configuration.txt')
mock1.pu_pstype_setpoint_limits.return_value = _read_test_file(
    'pwrsupply/putypes-setpoint-limits.txt')
mock1.ps_psmodels_read.return_value = _read_test_file(
    'pwrsupply/psmodels.txt')
mock1.pu_psmodels_read.return_value = _read_test_file(
    'pwrsupply/pumodels.txt')
mock1.beaglebone_freq_mapping.return_value = _read_test_file(
    'beaglebone/beaglebone-freq.txt')
mock1.bbb_udc_mapping.return_value = _read_test_file(
    'beaglebone/beaglebone-udc.txt')
mock1.udc_ps_mapping.return_value = _read_test_file('beaglebone/udc-bsmp.txt')

# --- mock objects ma_search _web ---
_p = mock.patch.object(ma_search, '_web', autospec=True)
mock2 = _p.start()
mock2.server_online.return_value = True
mock2.magnets_excitation_ps_read.return_value = _read_test_file(
    'magnet/magnet-excitation-ps.txt')
mock2.ps_pstypes_names_read.return_value = _read_test_file(
    'pwrsupply/pstypes-names.txt')
mock2.ps_pstype_data_read.side_effect = _ps_pstype_data_read
mock2.ps_pstype_setpoint_limits.return_value = _read_test_file(
    'pwrsupply/pstypes-setpoint-limits.txt')
# ps_mock_web.pu_pstype_setpoint_limits.return_value = _read_test_file(
#     'pwrsupply/putypes-setpoint-limits.txt')

# --- mock objects excdata _web ---
_p = mock.patch.object(excdata, '_web', autospec=True)
mock3 = _p.start()
mock3.magnets_excitation_data_read.side_effect = \
    _magnets_excitation_data_read

# --- mock objects hl_time_search _web ---
_p = mock.patch.object(hl_time_search, '_web', autospec=True)
mock4 = _p.start()
mock4.high_level_events.return_value = _read_test_file(
    'timesys/high-level-events.py')
mock4.high_level_triggers.return_value = _read_test_file(
    'timesys/high-level-triggers.py')

# --- mock objects ll_time_search _web ---
_p = mock.patch.object(ll_time_search, '_web', autospec=True)
mock5 = _p.start()
mock5.timing_devices_mapping.return_value = _read_test_file(
    'timesys/timing-devices-connection.txt')

# --- mock objects bpms_search _web ---
_p = mock.patch.object(bpms_search, '_web', autospec=True)
mock6 = _p.start()
mock6.bpms_data.return_value = _read_test_file(
    'diagnostics/bpms-data.txt')



"""State Simulator."""

import re as _re

from ..search import PSSearch as _PSSearch
from ..pwrsupply.csdev import get_ps_propty_database as _get_database
from ..pwrsupply.csdev import Const as _Const

from .simulator import Simulator as _Simulator


# NOTE: This is a toymodel PowerSupply simulator, just to test the
# Simulator infrastructure. Other PVs such as PwrState-* and
# OpMode-* should be added!

class SimPSTypeModel(_Simulator):
    """Power supply simulator."""

    # prefix for regexp used in pvname regexps
    # NOTE: this can be overriden in subclasses.
    _regexp_pvname = '.*:PS-.*'

    # these are PS properties whose correlations this simulator takes
    # into account.
    # NOTE: this can be overriden in subclasses.
    _properties_base = (
        'PwrState-Sel', 'PwrState-Sts',
        'OpMode-Sel', 'OpMode-Sts',
        )

    _properties = _properties_base + (
        'Current-SP', 'Current-RB',
        'CurrentRef-Mon', 'Current-Mon',
        # FBP-specific (added only if PS is of FBP pstype!)
        'SOFBCurrent-SP', 'SOFBCurrent-RB',
        'SOFBCurrentRef-Mon', 'SOFBCurrent-Mon',
        )

    # regexp used to determine setpoint PVs
    _regexp_setpoint = _regexp_setpoint = _re.compile('^.*-(SP|Sel|Cmd)$')

    _absolute_fluctuation = 0.001  # [A] or other subclass units

    def __init__(self, pstype, psmodel):
        """Initialize simulator.

        Static methd can be used to pass arguments.
        Ex.:
            typemodel = SimPSTypeModel.conv_psname2typemodel('BO-Fam:PS-QF')
            simps = SimPSTypeModel(*typemodel)
        """
        self._pstype = pstype
        self._psmodel = psmodel
        self._regexp_pv_check = _re.compile(self._regexp_pvname)

        # call base class constructor
        super().__init__()

        # list with all setpoint PVs in use (incremented in callback execution)
        self._setpoint_pvs = list()

    @property
    def pstype(self):
        """Return pstype."""
        return self._pstype

    @property
    def psmodel(self):
        """Return psmodel."""
        return self._psmodel

    def get_pwrstate(self, pvname):
        """Return PowerState associated with a given pvname."""
        pvn = pvname.substitute(propty='PwrState-Sts')
        if pvn not in self:
            # If PwrState-Sts not simulated, assume it to be On.
            return _Const.PwrStateSts.On
        return self.pv_value_get(pvn)

    @staticmethod
    def conv_psname2typemodel(psname):
        """Return pstype and psmodel associated with psname."""
        pstype = _PSSearch.conv_psname_2_pstype(psname)
        psmodel = _PSSearch.conv_psname_2_psmodel(psname)
        return pstype, psmodel

    def pv_check(self, pvname):
        """Check if SimPV belongs to simlutaor - overrides base method."""
        return self._regexp_pv_check.match(pvname)

    # --- base class abstract methods ---

    def callback_pv_dbase(self):
        """."""
        regexp_dbase = dict()
        dbpvs = _get_database(self._psmodel, self._pstype)
        for propty in self._properties:
            if propty in dbpvs:
                # NOTE: property is considered only if it is in dbase
                dbase = dbpvs[propty]
                regexp = self._regexp_pvname + ':' + propty
                regexp_dbase[regexp] = dbase
        return regexp_dbase

    def callback_pv_get(self, pvname, **kwargs):
        """Execute callback function prior to SimPV readout."""

    def callback_pv_put(self, pvname, value, **kwargs):
        """Execute callback setpoint to synchronize SimPVs."""
        # if not -SP pv, does not accept write by returning False
        if not SimPSTypeModel._regexp_setpoint.match(pvname):
            return False

        # synchronize other PVs
        status = self.callback_update(pvname=pvname, value=value, **kwargs)

        # NOTE: return status signaling wheter setpoint
        # was accepted and applied (True)
        return status

    def callback_pv_add(self, pvname):
        """."""
        # if of setpoint type, add pvname to list.
        if SimPSTypeModel._regexp_setpoint.match(pvname):
            self._setpoint_pvs.append(pvname)

    def callback_update(self, **kwargs):
        """Execute callback to update/synchronize simulator."""
        # get setpoints to be propagated in simulaltion
        if 'pvname' in kwargs:
            # this option is to optimize update callback invoked
            # from callback_set
            pvname = kwargs['pvname']
            setpoints = dict()
            setpoints[pvname] = kwargs['value']
        else:
            setpoints = dict()
            for sp_pvname in self._setpoint_pvs:
                setpoints[sp_pvname] = self.pv_value_get(sp_pvname)

        # synchronize Current PVs
        for sp_pvname, sp_value in setpoints.items():
            if '-SP' in sp_pvname:
                status = self._update_sp(sp_pvname, sp_value)
            elif '-Sel' in sp_pvname:
                status = self._update_sel(sp_pvname, sp_value)

        # return status
        return status

    # --- private methods ---

    def _update_sp(self, pvname, value):
        # -RB
        pvn = pvname.replace('-SP', '-RB')
        if pvn in self:
            self.pv_value_put(pvn, value)

        setpoint = value
        if self.get_pwrstate(pvname) != _Const.PwrStateSts.On:
            # if power supply is off, do not propagate further
            setpoint = 0 * setpoint

        # Ref-Mon
        pvn = pvname.replace('-SP', 'Ref-Mon')
        if pvn in self:
            self.pv_value_put(pvn, setpoint)

        # -Mon
        pvn = pvname.replace('-SP', '-Mon')
        if pvn in self:
            setpoint = super().Utils.add_fluctuations(
                setpoint,
                absolute=SimPSTypeModel._absolute_fluctuation)
            self.pv_value_put(pvn, setpoint)

        return True

    def _update_sel(self, pvname, value):
        pvn = pvname.replace('-Sel', '-Sts')
        if pvn in self:
            self.pv_value_put(pvn, value)

        return True


class SimPUTypeModel(SimPSTypeModel):
    """Pulsed power supply simulator."""

    _regexp_pvname = '.*:PU-.*'

    _properties = SimPSTypeModel._properties_base + (
        'Current-SP', 'Current-RB', 'Current-Mon',
        )



#!/usr/local/bin/python-sirius
"""Application module."""

import time as _time
from threading import Thread as _Thread

from ..envars import VACA_PREFIX as _VACA_PREFIX
from ..callbacks import Callback as _Callback
from ..thread import QueueThread as _QueueThread


class App(_Callback):
    """Main application responsible for updating DB."""

    SCAN_FREQUENCY = 2

    def __init__(self, *args):
        """Create Computed PVs."""
        super().__init__()
        self._prefix = _VACA_PREFIX
        self._queue = _QueueThread()
        self.pvs = list()
        self.scanning = False
        self.quit = False
        self._create_computed_pvs(*args)

        self.thread = _Thread(target=self.scan, daemon=True)
        self.thread.start()

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def read(self, reason):
        """Read from IOC database."""
        return None

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        return False  # return True to invoke super().write of PCASDriver

    def _create_computed_pvs(self):
        raise NotImplementedError

    def _update_pvs(self):
        raise NotImplementedError

    def scan(self):
        """Run as a thread scanning PVs."""
        while not self.quit:
            if self.scanning:
                self._update_pvs()
            _time.sleep(1.0/App.SCAN_FREQUENCY)



"""Main module of AS-AP-PosAng IOC."""

import logging as _log
import time as _time
import numpy as _np
from epics import PV as _PV

from .. import util as _util
from ..callbacks import Callback as _Callback
from ..envars import VACA_PREFIX as _vaca_prefix
from ..namesys import SiriusPVName as _SiriusPVName
from ..clientconfigdb import ConfigDBClient as _ConfigDBClient, \
    ConfigDBException as _ConfigDBException
from ..pwrsupply.csdev import Const as _PSC

from .csdev import Const as _PAConst, \
    get_posang_database as _get_database
from .utils import HandleConfigNameFile as _HandleConfigNameFile


# Constants
_TIMEOUT_CHECK = 3
_ALLSET = 0xf
_ALLCLR = 0x0


class App(_Callback):
    """Main application for handling injection in transport lines."""

    def __init__(self, tl, corrs_type):
        """Class constructor."""
        super().__init__()
        self._pvs_database = _get_database(tl, corrs_type)

        # consts
        self._TL = tl.upper()
        self._CORRSTYPE = corrs_type
        if self._TL == 'TS':
            corrh = (_PAConst.TS_CORRH_POSANG_CHSEPT
                     if self._CORRSTYPE == 'ch-sept'
                     else _PAConst.TS_CORRH_POSANG_SEPTSEPT)
            corrv = _PAConst.TS_CORRV_POSANG
        elif self._TL == 'TB':
            corrh = _PAConst.TB_CORRH_POSANG
            corrv = _PAConst.TB_CORRV_POSANG

        self._status = _ALLSET
        self._ref_check_update = _PAConst.NeedRefUpdate.NeedUpdate
        self._orbx_deltapos = 0
        self._orby_deltapos = 0
        self._orbx_deltaang = 0
        self._orby_deltaang = 0
        self._setnewrefkick_cmd_count = 0
        self._config_ps_cmd_count = 0

        self.cn_handler = _HandleConfigNameFile(self._TL, self._CORRSTYPE)
        self.cdb_client = _ConfigDBClient(
            config_type=self._TL.lower()+'_posang_respm')
        [done, corrparams] = self._get_corrparams()
        if done:
            self._config_name = corrparams[0]
            self._respmat_x = corrparams[1]
            self._respmat_y = corrparams[2]
        else:
            raise Exception(
                "Could not read correction parameters from configdb.")

        self._correctors = dict()
        self._correctors['CH1'] = _SiriusPVName(corrh[0])
        self._correctors['CH2'] = _SiriusPVName(corrh[1])
        if len(corrh) == 3:
            self._correctors['CH3'] = _SiriusPVName(corrh[2])
        self._correctors['CV1'] = _SiriusPVName(corrv[0])
        self._correctors['CV2'] = _SiriusPVName(corrv[1])
        if len(corrv) == 4:
            self._correctors['CV3'] = _SiriusPVName(corrv[2])
            self._correctors['CV4'] = _SiriusPVName(corrv[3])
        self._corrs2id = {v: k for k, v in self._correctors.items()}

        self._corr_check_connection = dict()
        self._corr_check_pwrstate_sts = dict()
        self._corr_check_opmode_sts = dict()
        self._corr_check_ctrlmode_mon = dict()
        # obs: ignore PU on OpMode and CtrlMode checks
        for corr_id, corr in self._correctors.items():
            self._corr_check_connection[corr_id] = 0
            self._corr_check_pwrstate_sts[corr_id] = 0
            if 'Sept' in corr.dev:
                self._corr_check_opmode_sts[corr_id] = _PSC.States.SlowRef
                self._corr_check_ctrlmode_mon[corr_id] = _PSC.Interface.Remote
            else:
                self._corr_check_opmode_sts[corr_id] = 1
                self._corr_check_ctrlmode_mon[corr_id] = 1

        # Connect to correctors
        self._corr_kick_sp_pvs = dict()
        self._corr_kick_rb_pvs = dict()
        self._corr_pwrstate_sel_pvs = dict()
        self._corr_pwrstate_sts_pvs = dict()
        self._corr_opmode_sel_pvs = dict()
        self._corr_opmode_sts_pvs = dict()
        self._corr_ctrlmode_mon_pvs = dict()
        self._corr_refkick = dict()
        self._corr_unit_factor = dict()
        self._corr_kick_rb_impl = dict()
        self._corr_last_delta = dict()
        self._corr_last_delta_impltd = dict()

        for corr in self._correctors.values():
            pss = corr.substitute(prefix=_vaca_prefix)
            self._corr_kick_sp_pvs[corr] = _PV(
                pss.substitute(propty_name='Kick', propty_suffix='SP'),
                connection_timeout=0.05)

            self._corr_refkick[corr] = 0
            self._corr_last_delta[corr] = 0
            self._corr_last_delta_impltd[corr] = True
            self._corr_kick_rb_impl[corr] = 0
            self._corr_kick_rb_pvs[corr] = _PV(
                pss.substitute(propty_name='Kick', propty_suffix='RB'),
                callback=[self._callback_init_refkick,
                          self._callback_corr_kick_rb],
                connection_callback=self._connection_callback_corr_kick_pvs,
                connection_timeout=0.05)

            self._corr_pwrstate_sel_pvs[corr] = _PV(
                pss.substitute(propty_name='PwrState', propty_suffix='Sel'),
                connection_timeout=0.05)
            self._corr_pwrstate_sts_pvs[corr] = _PV(
                pss.substitute(propty_name='PwrState', propty_suffix='Sts'),
                callback=self._callback_corr_pwrstate_sts,
                connection_timeout=0.05)

            if 'Sept' not in corr.dev:
                self._corr_opmode_sel_pvs[corr] = _PV(
                    pss.substitute(propty_name='OpMode', propty_suffix='Sel'),
                    connection_timeout=0.05)
                self._corr_opmode_sts_pvs[corr] = _PV(
                    pss.substitute(propty_name='OpMode', propty_suffix='Sts'),
                    callback=self._callback_corr_opmode_sts,
                    connection_timeout=0.05)

                self._corr_ctrlmode_mon_pvs[corr] = _PV(
                    pss.substitute(propty_name='CtrlMode',
                                   propty_suffix='Mon'),
                    callback=self._callback_corr_ctrlmode_mon,
                    connection_timeout=0.05)
                self._corr_unit_factor[corr] = 1e-6  # urad to rad
            else:
                self._corr_unit_factor[corr] = 1e-3  # mrad to rad

    def init_database(self):
        """Set initial PV values."""
        self.run_callbacks('ConfigName-SP', self._config_name)
        self.run_callbacks('ConfigName-RB', self._config_name)
        self.run_callbacks('RespMatX-Mon', self._respmat_x)
        self.run_callbacks('RespMatY-Mon', self._respmat_y)
        self.run_callbacks('Status-Mon', self._status)
        for corr_id, corr in self._correctors.items():
            self.run_callbacks(
                'RefKick' + corr_id + '-Mon', self._corr_refkick[corr])
        self._check_need_update_ref()
        msg = 'Started.'
        self.run_callbacks('Log-Mon', msg)
        _log.info('Log-Mon: ' + msg)

    @property
    def pvs_database(self):
        """Return pvs_database."""
        return self._pvs_database

    def process(self, interval):
        """Sleep."""
        _time.sleep(interval)

    def write(self, reason, value):
        """Write value to reason and let callback update PV database."""
        status = False
        if reason == 'DeltaPosX-SP':
            updated = self._update_delta(value, self._orbx_deltaang, 'x')
            if updated:
                self._orbx_deltapos = value
                self.run_callbacks('DeltaPosX-RB', value)
                status = True

        elif reason == 'DeltaAngX-SP':
            updated = self._update_delta(self._orbx_deltapos, value, 'x')
            if updated:
                self._orbx_deltaang = value
                self.run_callbacks('DeltaAngX-RB', value)
                status = True

        elif reason == 'DeltaPosY-SP':
            updated = self._update_delta(value, self._orby_deltaang, 'y')
            if updated:
                self._orby_deltapos = value
                self.run_callbacks('DeltaPosY-RB', value)
                status = True

        elif reason == 'DeltaAngY-SP':
            updated = self._update_delta(self._orby_deltapos, value, 'y')
            if updated:
                self._orby_deltaang = value
                self.run_callbacks('DeltaAngY-RB', value)
                status = True

        elif reason == 'SetNewRefKick-Cmd':
            updated = self._update_ref()
            if updated:
                self._setnewrefkick_cmd_count += 1
                self.run_callbacks(
                    'SetNewRefKick-Cmd', self._setnewrefkick_cmd_count)
                self._check_need_update_ref()

        elif reason == 'ConfigPS-Cmd':
            done = self._config_ps()
            if done:
                self._config_ps_cmd_count += 1
                self.run_callbacks('ConfigPS-Cmd', self._config_ps_cmd_count)

        elif reason == 'ConfigName-SP':
            [done, corrparams] = self._get_corrparams(value)
            if done:
                self.cn_handler.set_config_name(value)
                self._config_name = corrparams[0]
                self.run_callbacks('ConfigName-RB', self._config_name)
                self._respmat_x = corrparams[1]
                self.run_callbacks('RespMatX-Mon', self._respmat_x)
                self._respmat_y = corrparams[2]
                self.run_callbacks('RespMatY-Mon', self._respmat_y)
                self._update_delta(
                    self._orbx_deltapos, self._orbx_deltaang, 'x')
                self._update_delta(
                    self._orby_deltapos, self._orby_deltaang, 'y')
                msg = 'Updated correction matrices.'
                status = True
            else:
                msg = 'ERR:Configuration not found in configdb.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)

        return status  # return True to invoke super().write of PCASDriver

    def _get_corrparams(self, config_name=''):
        """Get response matrix from configurations database."""
        try:
            if not config_name:
                config_name = self.cn_handler.get_config_name()
            mats = self.cdb_client.get_config_value(config_name)
        except _ConfigDBException:
            return [False, []]

        respmat_x = [item for sublist in mats['respm-x'] for item in sublist]
        respmat_y = [item for sublist in mats['respm-y'] for item in sublist]
        return [True, [config_name, respmat_x, respmat_y]]

    def _update_delta(self, delta_pos, delta_ang, orbit):
        corrs2delta = list()
        if orbit == 'x':
            respmat = self._respmat_x
            corrs2delta.append((self._correctors['CH1'], 0))
            if 'CH3' in self._correctors.keys():
                corrs2delta.append((self._correctors['CH2'], 0))
                corrs2delta.append((self._correctors['CH3'], 1))
            else:
                corrs2delta.append((self._correctors['CH2'], 1))
        else:
            respmat = self._respmat_y
            corrs2delta.append((self._correctors['CV1'], 0))
            corrs2delta.append((self._correctors['CV2'], 1))
            if 'CV3' in self._correctors.keys():
                corrs2delta.append((self._correctors['CV3'], 2))
                corrs2delta.append((self._correctors['CV4'], 3))

        if self._status != _ALLCLR:
            msg = 'ERR:Failed on applying new delta.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)
            return False

        # Convert deltas to respm units: pos mm->m and ang mrad->rad
        deltaposang = _np.array([[delta_pos*1e-3], [delta_ang*1e-3]])
        mat = _np.reshape(respmat, (2, -1), order='C')
        try:
            umat, smat, vmat = _np.linalg.svd(mat, full_matrices=False)
            invmat = _np.dot(_np.dot(vmat.T, _np.diag(1/smat)), umat.T)
        except _np.linalg.LinAlgError():
            msg = 'ERR: Could not calculate SVD.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)
            return False
        if _np.any(_np.isnan(invmat)) or _np.any(_np.isinf(invmat)):
            msg = 'ERR: Pseudo inverse contains nan or inf.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)
            return False
        deltas = _np.dot(invmat, deltaposang)

        # Convert kicks from rad to correctors units and send values
        sp_check = dict()
        for corr, idx in corrs2delta:
            # delta from rad to urad or mrad
            dlt = deltas[idx][0]/self._corr_unit_factor[corr]
            self._corr_last_delta[corr] = dlt
            val = self._corr_refkick[corr] + dlt
            self._corr_kick_sp_pvs[corr].put(val)
            sp_check.update({corr: [False, val]})

        # check if SP were accepted
        time0 = _time.time()
        while _time.time() - time0 < _TIMEOUT_CHECK:
            for corr in sp_check:
                if sp_check[corr][0]:
                    continue
                desired = sp_check[corr][1]
                currval = self._corr_kick_rb_pvs[corr].get()
                isok = currval is not None and \
                    _np.isclose(currval, desired)
                sp_check[corr] = [isok, desired]
            if not any(not val[0] for val in sp_check.values()):
                break

        sp_diff = False
        for corr in sp_check:
            self._corr_last_delta_impltd[corr] = sp_check[corr][0]
            if not sp_check[corr][0]:
                msg = 'ERR: Delta not applied to '+corr+'.'
                self.run_callbacks('Log-Mon', msg)
                _log.info('Log-Mon: ' + msg)
                sp_diff = True
        if sp_diff:
            return False
        msg = 'Applied new delta.'
        self.run_callbacks('Log-Mon', msg)
        _log.info('Log-Mon: ' + msg)
        return True

    def _update_ref(self):
        if (self._status & 0x1) == 0:  # Check connection
            # updates reference
            for corr_id, corr in self._correctors.items():
                value = self._corr_kick_rb_pvs[corr].get()
                # Get correctors kick in urad (PS) or mrad (PU).
                self._corr_refkick[corr] = value
                self._corr_last_delta[corr] = 0.0
                self.run_callbacks('RefKick' + corr_id + '-Mon', value)

            # the deltas from new kick references are zero
            self._orbx_deltapos = 0
            self.run_callbacks('DeltaPosX-SP', 0)
            self.run_callbacks('DeltaPosX-RB', 0)
            self._orbx_deltaang = 0
            self.run_callbacks('DeltaAngX-SP', 0)
            self.run_callbacks('DeltaAngX-RB', 0)
            self._orby_deltapos = 0
            self.run_callbacks('DeltaPosY-SP', 0)
            self.run_callbacks('DeltaPosY-RB', 0)
            self._orby_deltaang = 0
            self.run_callbacks('DeltaAngY-SP', 0)
            self.run_callbacks('DeltaAngY-RB', 0)

            msg = 'Updated Kick References.'
            updated = True
        else:
            msg = 'ERR:Some pv is disconnected.'
            updated = False
        self.run_callbacks('Log-Mon', msg)
        _log.info('Log-Mon: ' + msg)
        return updated

    def _callback_init_refkick(self, pvname, value, cb_info, **kws):
        """Initialize RefKick-Mon pvs and remove this callback."""
        if value is None:
            return
        corr = _SiriusPVName(pvname).device_name
        corr_id = self._corrs2id[corr]

        # Get reference. Correctors kick in urad (PS) or mrad (PU).
        self._corr_refkick[corr] = value
        self.run_callbacks(
            'RefKick' + corr_id + '-Mon', self._corr_refkick[corr])

        # Remove callback
        cb_info[1].remove_callback(cb_info[0])
        self._check_need_update_ref()

    def _connection_callback_corr_kick_pvs(self, pvname, conn, **kws):
        if not conn:
            msg = 'WARN:'+pvname+' disconnected.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)

        corr_id = self._corrs2id[_SiriusPVName(pvname).device_name]
        self._corr_check_connection[corr_id] = (1 if conn else 0)

        # Change the first bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=0,
            bit_val=any(q == 0 for q in self._corr_check_connection.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_corr_pwrstate_sts(self, pvname, value, **kws):
        if value != _PSC.PwrStateSts.On:
            msg = 'WARN:'+pvname+' is not On.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)

        corr_id = self._corrs2id[_SiriusPVName(pvname).device_name]
        self._corr_check_pwrstate_sts[corr_id] = value

        # Change the second bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=1,
            bit_val=any(q != _PSC.PwrStateSts.On
                        for q in self._corr_check_pwrstate_sts.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_corr_opmode_sts(self, pvname, value, **kws):
        msg = 'WARN:'+pvname+' changed.'
        self.run_callbacks('Log-Mon', msg)
        _log.info('Log-Mon: ' + msg)

        corr_id = self._corrs2id[_SiriusPVName(pvname).device_name]
        self._corr_check_opmode_sts[corr_id] = value

        # Change the third bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=2,
            bit_val=any(s != _PSC.States.SlowRef
                        for s in self._corr_check_opmode_sts.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_corr_ctrlmode_mon(self, pvname, value, **kws):
        if value != _PSC.Interface.Remote:
            msg = 'WARN:'+pvname+' is not Remote.'
            self.run_callbacks('Log-Mon', msg)
            _log.info('Log-Mon: ' + msg)

        corr_id = self._corrs2id[_SiriusPVName(pvname).device_name]
        self._corr_check_ctrlmode_mon[corr_id] = value

        # Change the fourth bit of correction status
        self._status = _util.update_bit(
            v=self._status, bit_pos=3,
            bit_val=any(q != _PSC.Interface.Remote
                        for q in self._corr_check_ctrlmode_mon.values()))
        self.run_callbacks('Status-Mon', self._status)

    def _callback_corr_kick_rb(self, pvname, value, **kws):
        corr = _SiriusPVName(pvname).device_name
        self._corr_kick_rb_impl[corr] = value
        self._check_need_update_ref()

    def _check_need_update_ref(self):
        self._ref_check_update = _PAConst.NeedRefUpdate.Ok
        for corr in self._correctors.values():
            if not self._corr_last_delta_impltd[corr]:
                continue
            implemented = self._corr_kick_rb_impl[corr]
            desired = self._corr_last_delta[corr] + self._corr_refkick[corr]
            if not _np.isclose(implemented, desired):
                self._ref_check_update = _PAConst.NeedRefUpdate.NeedUpdate
                break
        self.run_callbacks('NeedRefUpdate-Mon', self._ref_check_update)

    def _config_ps(self):
        for corr in self._correctors.values():
            if self._corr_pwrstate_sel_pvs[corr].connected:
                self._corr_pwrstate_sel_pvs[corr].put(_PSC.PwrStateSel.On)
                if 'Sept' not in corr:
                    self._corr_opmode_sel_pvs[corr].put(_PSC.OpMode.SlowRef)
            else:
                msg = 'ERR:'+corr+' is disconnected.'
                self.run_callbacks('Log-Mon', msg)
                _log.info('Log-Mon: ' + msg)
                return False
        msg = 'Configuration sent to correctors.'
        self.run_callbacks('Log-Mon', msg)
        _log.info('Log-Mon: ' + msg)
        return True



"""BO orbit correction response matrix configuration.

Values in _template_dict are arbitrary. They are used just to compare with
corresponding values when a new configuration is tried to be inserted in the
servconf database.
"""
from copy import deepcopy as _dcopy


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template),
        'check': False,
    }
    return _dict


# Orbit Correction Response Matrix for Booster
#   -First bpm in the list is the first seen by the beam during injection
#   -First ch in the list is the first seen by the beam during injection
#   -First cv in the list is the first seen by the beam during injection
#   -RF is the last component;
#   -Units: bpm --> um; (ch, cv) --> urad; RF --> Hz
#
# | BPMXi |   | Mik ... Mil ... Min |   | CHk |
# |  ...  | = | ... ... ... ... ... | * | ... |
# | BPMYj |   | Mjk ... Mjl ... Mjn |   | CVl |
#                                       | ... |
#                                       | RFn |

_template = 100*[51*[0.0]]



"""Parameters of the IOCs of the AP discipline."""
from copy import deepcopy as _dcopy

_off = 0
_on = 1


def get_dict():
    """Return configuration type dictionary."""
    module_name = __name__.split('.')[-1]
    _dict = {
        'config_type_name': module_name,
        'value': _dcopy(_template_dict),
        'check': False,
    }
    return _dict


# When using this type of configuration to set the machine,
# the list of PVs should be processed in the same order they are stored
# in the configuration. The second numeric parameter in the pair is the
# delay [s] the client should wait before setting the next PV.

_tb_sofb = [
    ['TB-Glob:AP-SOFB:BPMXEnblList-SP', [0, ]*30, 0.0],
    ['TB-Glob:AP-SOFB:BPMYEnblList-SP', [0, ]*30, 0.0],
    ['TB-Glob:AP-SOFB:CHEnblList-SP', [0, ]*5, 0.0],
    ['TB-Glob:AP-SOFB:CVEnblList-SP', [0, ]*6, 0.0],
    ['TB-Glob:AP-SOFB:ManCorrGainCH-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:ManCorrGainCV-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:KickAcqRate-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MaxDeltaKickCH-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MaxDeltaKickCV-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MaxKickCH-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MaxKickCV-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MeasRespMatKickCH-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MeasRespMatKickCV-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:MeasRespMatWait-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:OfflineOrbX-SP', [0.0, ]*30, 0.0],
    ['TB-Glob:AP-SOFB:OfflineOrbY-SP', [0.0, ]*30, 0.0],
    ['TB-Glob:AP-SOFB:OrbAcqRate-SP', 0.0, 0.0],
    ['TB-Glob:AP-SOFB:RefOrbX-SP', [0.0, ]*30, 0.0],
    ['TB-Glob:AP-SOFB:RefOrbY-SP', [0.0, ]*30, 0.0],
    ['TB-Glob:AP-SOFB:RespMat-SP', [0.0, ]*660, 0.0],
    ['TB-Glob:AP-SOFB:RespMatMode-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:SOFBMode-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:SPassAvgNrTurns-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:SPassMaskSplBeg-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:SPassMaskSplEnd-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:SPassMethod-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:SPassUseBg-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:SmoothMethod-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:SmoothNrPts-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigAcqChan-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigAcqCtrl-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigAcqRepeat-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigAcqTrigger-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigDataChan-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigDataHyst-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigDataPol-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigDataSel-Sel', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigDataThres-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigNrSamplesPost-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigNrSamplesPre-SP', 0, 0.0],
    ['TB-Glob:AP-SOFB:TrigNrShots-SP', 0, 0.0],
]

_bo_sofb = [
    ['BO-Glob:AP-SOFB:BPMXEnblList-SP', [0, ]*250, 0.0],
    ['BO-Glob:AP-SOFB:BPMYEnblList-SP', [0, ]*250, 0.0],
    ['BO-Glob:AP-SOFB:CHEnblList-SP', [0, ]*25, 0.0],
    ['BO-Glob:AP-SOFB:CVEnblList-SP', [0, ]*25, 0.0],
    ['BO-Glob:AP-SOFB:LoopFreq-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:CorrSync-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:ManCorrGainCH-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:ManCorrGainCV-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:KickAcqRate-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MTurnDownSample-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:MTurnIdx-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:MTurnMaskSplBeg-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:MTurnMaskSplEnd-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:MTurnSyncTim-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:MTurnUseMask-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:MaxDeltaKickCH-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MaxDeltaKickCV-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MaxKickCH-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MaxKickCV-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MeasRespMatKickCH-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MeasRespMatKickCV-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:MeasRespMatWait-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:OfflineOrbX-SP', [0.0, ]*250, 0.0],
    ['BO-Glob:AP-SOFB:OfflineOrbY-SP', [0.0, ]*250, 0.0],
    ['BO-Glob:AP-SOFB:OrbAcqRate-SP', 0.0, 0.0],
    ['BO-Glob:AP-SOFB:RFEnbl-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:RefOrbX-SP', [0.0, ]*250, 0.0],
    ['BO-Glob:AP-SOFB:RefOrbY-SP', [0.0, ]*250, 0.0],
    ['BO-Glob:AP-SOFB:RespMat-SP', [0.0, ]*25500, 0.0],
    ['BO-Glob:AP-SOFB:RespMatMode-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:RingSize-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:SOFBMode-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:SPassAvgNrTurns-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:SPassMaskSplBeg-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:SPassMaskSplEnd-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:SPassMethod-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:SPassUseBg-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:SmoothMethod-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:SmoothNrPts-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigAcqChan-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigAcqCtrl-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigAcqRepeat-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigAcqTrigger-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigDataChan-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigDataHyst-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigDataPol-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigDataSel-Sel', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigDataThres-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigNrSamplesPost-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigNrSamplesPre-SP', 0, 0.0],
    ['BO-Glob:AP-SOFB:TrigNrShots-SP', 0, 0.0],
]

_ts_sofb = [
    ['TS-Glob:AP-SOFB:BPMXEnblList-SP', [0, ]*25, 0.0],
    ['TS-Glob:AP-SOFB:BPMYEnblList-SP', [0, ]*25, 0.0],
    ['TS-Glob:AP-SOFB:CHEnblList-SP', [0, ]*4, 0.0],
    ['TS-Glob:AP-SOFB:CVEnblList-SP', [0, ]*6, 0.0],
    ['TS-Glob:AP-SOFB:ManCorrGainCH-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:ManCorrGainCV-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:KickAcqRate-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MaxDeltaKickCH-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MaxDeltaKickCV-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MaxKickCH-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MaxKickCV-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MeasRespMatKickCH-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MeasRespMatKickCV-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:MeasRespMatWait-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:OfflineOrbX-SP', [0.0, ]*25, 0.0],
    ['TS-Glob:AP-SOFB:OfflineOrbY-SP', [0.0, ]*25, 0.0],
    ['TS-Glob:AP-SOFB:OrbAcqRate-SP', 0.0, 0.0],
    ['TS-Glob:AP-SOFB:RefOrbX-SP', [0.0, ]*25, 0.0],
    ['TS-Glob:AP-SOFB:RefOrbY-SP', [0.0, ]*25, 0.0],
    ['TS-Glob:AP-SOFB:RespMat-SP', [0.0, ]*500, 0.0],
    ['TS-Glob:AP-SOFB:RespMatMode-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:SOFBMode-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:SPassAvgNrTurns-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:SPassMaskSplBeg-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:SPassMaskSplEnd-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:SPassMethod-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:SPassUseBg-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:SmoothMethod-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:SmoothNrPts-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigAcqChan-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigAcqCtrl-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigAcqRepeat-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigAcqTrigger-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigDataChan-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigDataHyst-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigDataPol-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigDataSel-Sel', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigDataThres-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigNrSamplesPost-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigNrSamplesPre-SP', 0, 0.0],
    ['TS-Glob:AP-SOFB:TrigNrShots-SP', 0, 0.0],
]

_si_sofb = [
    ['SI-Glob:AP-SOFB:BPMXEnblList-SP', [0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:BPMYEnblList-SP', [0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:CHEnblList-SP', [0, ]*120, 0.0],
    ['SI-Glob:AP-SOFB:CVEnblList-SP', [0, ]*160, 0.0],
    ['SI-Glob:AP-SOFB:LoopFreq-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:CorrSync-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:ManCorrGainCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:ManCorrGainCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:ManCorrGainRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:KickAcqRate-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MTurnDownSample-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:MTurnIdx-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:MTurnMaskSplBeg-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:MTurnMaskSplEnd-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:MTurnSyncTim-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:MTurnUseMask-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:MaxDeltaKickCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxDeltaKickCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxDeltaKickRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxKickCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MaxKickCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MeasRespMatKickCH-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MeasRespMatKickCV-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MeasRespMatKickRF-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:MeasRespMatWait-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:OfflineOrbX-SP', [0.0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:OfflineOrbY-SP', [0.0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:OrbAcqRate-SP', 0.0, 0.0],
    ['SI-Glob:AP-SOFB:RFEnbl-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:RefOrbX-SP', [0.0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:RefOrbY-SP', [0.0, ]*800, 0.0],
    ['SI-Glob:AP-SOFB:RespMat-SP', [0.0, ]*449600, 0.0],
    ['SI-Glob:AP-SOFB:RespMatMode-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:RingSize-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:SOFBMode-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:SPassAvgNrTurns-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:SPassMaskSplBeg-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:SPassMaskSplEnd-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:SPassMethod-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:SPassUseBg-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:SmoothMethod-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:SmoothNrPts-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigAcqChan-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigAcqCtrl-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigAcqRepeat-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigAcqTrigger-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigDataChan-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigDataHyst-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigDataPol-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigDataSel-Sel', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigDataThres-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigNrSamplesPost-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigNrSamplesPre-SP', 0, 0.0],
    ['SI-Glob:AP-SOFB:TrigNrShots-SP', 0, 0.0],
]

_sofb = _tb_sofb + _bo_sofb + _ts_sofb + _si_sofb

_template_dict = {'pvs': _sofb}



"""Power Supply Signal Generator."""

import time as _t
import math as _math


DEFAULT_SIGGEN_CONFIG = (
    0,      # sigtype  [0:Sine]
    100,    # num_cycles
    2.0,    # freq [Hz]
    0.0,    # amplitude [A] (Maximum amplitude)
    0.0,    # offset [A]
    0.0,    # aux_param[0]
            #   (Sine|DampedSine|DampedSquaredSine: theta_beg,
            #    Trapezoidal: ramp up[s])
    0.0,    # aux_param[1]
            #   (Sine|DampedSine|DampedSquaredSine: theta_end,
            #    Trapezoidal: ramp down[s])
    0.0,    # aux_param[2]
            #   (DampedSine|DampedSquaredSine: decay time [s])
    0.0     # aux_param[3]
            #   (reserved)
    )


class Signal:
    """Signal from SigGen."""

    def __init__(self,
                 sigtype,
                 num_cycles,  # Sine, DampedSine, DampedSqrdSin, Trapezoidal
                 freq,  # [Hz] Sine, DampedSine, DampedSqrdSin
                 amplitude,  # [A] Sine, DampedSine, DampedSqrdSin, Trapezoidal
                 offset,  # [A] Sine, DampedSine, DampedSqrdSin, Trapezoidal
                 aux_param,  # Sine, DampedSine, DampedSqrdSin, Trapezoidal)
                 **kwargs
                 ):
        """Init method."""
        _ = kwargs  # throwaway arguments
        self.sigtype = sigtype
        self.num_cycles = num_cycles
        self.freq = freq
        self.amplitude = amplitude
        self.offset = offset
        self.aux_param = aux_param
        self.enable = True
        self.time_init = _t.time()
        self.time_delta = self.time_init

    @property
    def duration(self):
        """Return total duration. Zero is infinite time."""
        return self._get_duration()

    @property
    def value(self):
        """Return current signal value."""
        if self.enable:
            self.time_delta = _t.time() - self.time_init
        return self._get_value(self.time_delta)

    @property
    def cycle_time(self):
        """Return period of signal."""
        return self._get_cycle_time()

    @property
    def rampup_time(self):
        """Rampup time for Trapezoidal signals."""
        return self.aux_param[0]

    @rampup_time.setter
    def rampup_time(self, value):
        """Set Rampup time for Trapezoidal signals."""
        self.aux_param[0] = value
        self._update()
        return value

    @property
    def theta_begin(self):
        """Return initial phase for Sine or Damped(Squared)Sine signals."""
        return self.aux_param[0]

    @theta_begin.setter
    def theta_begin(self, value):
        """Set Initial phase for Sine or Damped(Squared)Sine signals."""
        self.aux_param[0] = value
        self._update()
        return value

    @property
    def rampdown_time(self):
        """Rampdown time for Trapezoidal signals."""
        return self.aux_param[2]

    @rampdown_time.setter
    def rampdown_time(self, value):
        """Set Rampdown time for Trapezoidal signals."""
        self.aux_param[2] = value
        self._update()
        return value

    @property
    def theta_end(self):
        """Return final phase for Sine or Damped(Squared)Sine signals."""
        return self.aux_param[1]

    @theta_end.setter
    def theta_end(self, value):
        """Set Final phase for Sine or Damped(Squared)Sine signals."""
        self.aux_param[1] = value
        self._update()
        return value

    @property
    def plateau_time(self):
        """Plateau time for Trapezoidal signals."""
        return self.aux_param[1]

    @plateau_time.setter
    def plateau_time(self, value):
        """Set Plateau time for Trapezoidal signals."""
        self.aux_param[1] = value
        self._update()
        return value

    @property
    def decay_time(self):
        """Decay time constant for Damped(Squared)Sine signals."""
        return self.aux_param[2]

    @decay_time.setter
    def decay_time(self, value):
        """Set Decay time constant for Damped(Squared)Sine signals."""
        self.aux_param[2] = value
        self._update()
        return value

    def reset(self):
        """Reset init time."""
        self.time_init = _t.time()

    def get_waveform(self, nrpts=100):
        """."""
        tmax = self.duration
        tstep = tmax/(nrpts-1) if nrpts > 1 else tmax
        tval, wval = [], []
        time = 0
        while True:
            wvalue = self._get_value(time)
            _ = tval.append(time), wval.append(wvalue)
            if len(tval) == nrpts:
                break
            time += tstep
            if time > tmax:
                time = tmax
        return wval, tval

    # --- virtual methods ---

    def _get_duration(self):
        raise NotImplementedError

    def _get_cycle_time(self):
        raise NotImplementedError

    def _get_value(self, time_delta):
        raise NotImplementedError

    def _update(self):
        raise NotImplementedError


class SignalSine(Signal):
    """Sine signal."""

    def _get_duration(self):
        return self.num_cycles / self.freq

    def _get_cycle_time(self):
        return 1.0 / self.freq

    def _get_value(self, time_delta):
        if self.duration > 0 and time_delta > self.duration:
            # self.enable = False
            return self.offset
        else:
            value = self.offset + self.amplitude * \
                self._get_sin_signal(time_delta)
            return value

    def _get_sin_signal(self, time_delta):
        value = _math.sin(2 * _math.pi * self.freq * time_delta +
                          _math.radians(self.theta_begin))
        return value

    def _update(self):
        pass


class SignalDampedNSine(SignalSine):
    """DampedNSine signal."""

    def __init__(self, n, **kwargs):
        """Init method."""
        super().__init__(**kwargs)
        self.n = n
        self._update()

    def _get_sin_signal(self, time_delta):
        sinsig = (super()._get_sin_signal(time_delta))**self.n
        expsig = self._f * _math.exp(-time_delta/self.decay_time)
        value = sinsig * expsig
        return value

    def _update(self):
        self.wfreq = 2*_math.pi*self.freq
        if self.wfreq != 0.0:
            self._t0 = _math.atan(self.wfreq*self.decay_time*self.n)/self.wfreq
        else:
            self._t0 = 0.0
        self._f = _math.exp(self._t0/self.decay_time) / \
            _math.sin(self.wfreq*self._t0)**self.n


class SignalDampedSine(SignalDampedNSine):
    """DampedSine signal."""

    def __init__(self, **kwargs):
        """Init method."""
        super().__init__(n=1, **kwargs)


class SignalDampedSquaredSine(SignalDampedNSine):
    """DampedSquaredSine signal."""

    def __init__(self, **kwargs):
        """Init method."""
        super().__init__(n=2, **kwargs)


class SignalTrapezoidal(Signal):
    """Trapezoidal signal."""

    def _get_duration(self):
        return self.cycle_time * self.num_cycles

    def _get_cycle_time(self):
        return self.rampup_time + self.plateau_time + self.rampdown_time

    def _get_value(self, time_delta):
        if self.duration > 0 and time_delta > self.duration:
            # self.enable = False
            return self.offset
        else:
            cycle_pos = time_delta % self.cycle_time
            if cycle_pos < self.rampup_time:
                value = self.amplitude*cycle_pos/self.rampup_time
            elif cycle_pos < (self.rampup_time + self.plateau_time):
                value = self.amplitude
            else:
                down_time = cycle_pos - (self.rampup_time + self.plateau_time)
                value = self.amplitude*(1 - down_time/self.rampdown_time)
            return self.offset + value

    def _check(self):
        # TODO: avoid this workaround!
        if self.rampup_time == 0:
            self.rampup_time = 0.1
        if self.plateau_time == 0:
            self.plateau_time = 0.1
        if self.rampdown_time == 0:
            self.rampdown_time = 0.1

    def _update(self):
        pass


class SignalSquare(SignalSine):
    """Square signal."""

    def _get_value(self, time_delta):
        if self.duration > 0 and time_delta > self.duration:
            value = self.offset
        elif self._get_sin_signal(time_delta) < 0:
            value = - self.amplitude + self.offset
        else:
            value = self.amplitude + self.offset
        return value


class SigGenFactory:
    """Signal Generator Factory."""

    TYPES = {
        'Sine': 0,
        'DampedSine': 1,
        'Trapezoidal': 2,
        'DampedSquaredSine': 3,
        'Square': 4}

    TYPES_IND = {v: k for k, v in TYPES.items()}

    DEFAULT_CONFIGS = {
        'Sine': DEFAULT_SIGGEN_CONFIG,
        'DampedSine': [1, 1, 100.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
        'Trapezoidal': [2, 1, 0.0, 0.0, 0.0, 0.01, 0.01, 0.01, 0.0],
        'DampedSquaredSine': [3, 1, 100.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0],
        'Square': [4, 3.0, 0.1, 5.0, 5.0, 270.0, 270.0, 0.0, 0.0],
        }

    @staticmethod
    def create(data=None, **kwargs):
        """Implement factory method.

            This methods returns a Signal object corresponding to the
        selected type.

        Valid arguments:

        sigtype: Signal type, int or str (Sine|DampedSine|Trapezoidal|Square)
        num_cycles: Number of cycles, int (Sine|DampedSine|Trapezoidal|Square)
        freq: Frequency [Hz], float, (Sine|DampedSine|Square)
        amplitude: Amplitude [A], float, (Sine|DampedSine|Trapezoidal|Square)
        offset: Offset [A], float, (Sine|DampedSine|Trapezoidal|Square)
        aux_param: Aux. Parameters, float4 (Sine|DampedSine|Trapezoidal|Square)
        rampup_time: Rampup time [s], float, (Trapezoidal) - aux_param[0]
        rampdown_time: Rampdown time [s], float, (Trapezoidal) - aux_param[1]
        plateau_time: Plateau time [s], float, (Trapezoidal) - aux_param[2]
        theta_begin: Initial phase[deg] (Sine|DampedSine|Square) - aux_param[0]
        theta_end: Final phase[deg] (Sine|DampedSine|Square) - aux_param[1]
        decay_time: Decay time [s] (DampedSine) - aux_param[2]
        """
        # set signal type
        if 'sigtype' in kwargs:
            sigtype = kwargs['sigtype']
        elif data is not None:
            sigtype = data[0]
        else:
            sigtype = SigGenFactory.TYPES['Sine']
        if isinstance(sigtype, str):
            sigtype = SigGenFactory.TYPES[sigtype]

        # set ps controller initial values
        kwa = SigGenFactory._set_kwa(kwargs, sigtype, data)

        if sigtype == SigGenFactory.TYPES['Trapezoidal']:
            return SignalTrapezoidal(**kwa)
        elif sigtype == SigGenFactory.TYPES['Sine']:
            return SignalSine(**kwa)
        elif sigtype == SigGenFactory.TYPES['DampedSine']:
            return SignalDampedSine(**kwa)
        elif sigtype == SigGenFactory.TYPES['DampedSquaredSine']:
            return SignalDampedSquaredSine(**kwa)
        elif sigtype == SigGenFactory.TYPES['Square']:
            return SignalSquare(**kwa)

        # NOTE: this point should not be reached!
        return None

    @staticmethod
    def _set_kwa(kwargs, sigtype, data):
        # set ps controller initial values
        kwa = dict()
        kwa.update(kwargs)
        kwa['sigtype'] = sigtype
        parms = SigGenFactory.DEFAULT_CONFIGS[SigGenFactory.TYPES_IND[sigtype]]

        kwa['num_cycles'] = parms[1]
        kwa['freq'] = parms[2]  # [A]
        kwa['amplitude'] = parms[3]
        kwa['offset'] = parms[4]  # [A]
        kwa['aux_param'] = parms[5:9]

        # process data argument
        # kwa = dict()
        if data is not None:
            kwa['sigtype'] = data[0]
            kwa['num_cycles'] = int(data[1])
            kwa['freq'] = float(data[2])
            kwa['amplitude'] = float(data[3])
            kwa['offset'] = float(data[4])
            kwa['aux_param'] = [float(d) for d in data[5:9]]

        # process type argument
        kwa.update(kwargs)
        if 'rampup_time' in kwa:
            kwa['aux_param'][0] = float(kwa['rampup_time'])
        if 'rampdown_time' in kwa:
            kwa['aux_param'][1] = float(kwa['rampdown_time'])
        if 'plateau_time' in kwa:
            kwa['aux_param'][2] = float(kwa['plateau_time'])
        if 'theta_begin' in kwa:
            kwa['aux_param'][0] = float(kwa['theta_begin'])
        if 'theta_end' in kwa:
            kwa['aux_param'][1] = float(kwa['theta_end'])
        if 'decay_time' in kwa:
            kwa['aux_param'][2] = float(kwa['decay_time'])

        return kwa



#!/usr/bin/env python-sirius

"""Unittest module for envars.py."""

from unittest import TestCase
import siriuspy.envars as envars
import siriuspy.util as util

DIRS = (
    'DIR_ROOT',
    'DIR_EPICS_BASE',
    'DIR_SIRIUS',
    'DIR_FACS',
    'DIR_IMAS',
    'DIR_SIRIUS_CODE',
    'DIR_FACS_CODE',
    'DIR_SIRIUS_CODE_CSCNSTS',
    'DIR_SIRIUS_CODE_SIRIUSPY',
    'DIR_SIRIUS_CODE_HLA')
SRVURLS = (
    'SRVURL_RBACAUTH',
    'SRVURL_RBAC',
    'SRVURL_NS',
    'SRVURL_CCDB',
    'SRVURL_CABLES',
    'SRVURL_CSCONSTS',
    'SRVURL_CSCONSTS_2',
    'SRVURL_LOGBOOK',
    'SRVURL_CONFIGDB',
    'SRVURL_CONFIGDB_2',
    'SRVURL_ARCHIVER')
PUB_INTERFACE = DIRS + SRVURLS + ('VACA_PREFIX', )


class TestEnvars(TestCase):
    """Test envars."""

    def test_public_interface(self):
        """Test module's public interface."""
        valid = util.check_public_interface_namespace(envars, PUB_INTERFACE)
        self.assertTrue(valid)

    def test_dirs(self):
        """Test folder names."""
        for folder in DIRS:
            self.assertIn(folder, envars.__dict__)
            value = getattr(envars, folder)
            self.assertIsInstance(value, str)

    def test_srvurls(self):
        """Test server names."""
        for server in SRVURLS:
            self.assertIn(server, envars.__dict__)
            value = getattr(envars, server)
            self.assertIsInstance(value, str)

    def test_vaca(self):
        """VACA prefix string."""
        self.assertIn('VACA_PREFIX', envars.__dict__)
        value = getattr(envars, 'VACA_PREFIX')
        self.assertIsInstance(value, str)



#!/usr/bin/env python-sirius

"""Test entities module."""

import struct
from unittest import TestCase

from siriuspy.bsmp import Function, Types, Variable, VariablesGroup
from siriuspy.util import check_public_interface_namespace


class TestBSMPVariable(TestCase):
    """Test Variable class."""

    api = ('load_to_value', 'value_to_load')

    def test_api(self):
        """Test API."""
        self.assertTrue(check_public_interface_namespace(Variable, self.api))

    def test_int_value_to_load(self):
        """Test value to load method."""
        value = 10
        expected_load = [chr(b) for b in struct.pack('<b', 10)]
        var = Variable(1, False, Types.T_UINT8, 1)
        self.assertEqual(var.value_to_load(value), expected_load)

    def test_float_value_to_load(self):
        """Test value to load method."""
        value = 10.5
        expected_load = [chr(b) for b in struct.pack('<f', value)]
        var = Variable(1, False, Types.T_FLOAT, 1)
        self.assertEqual(var.value_to_load(value), expected_load)

    def test_string_value_to_load(self):
        """Test value to load method."""
        value = 'V0.07 2018-03-14V0.07 2018-03-14'

        expected_load = ['V', '0', '.', '0', '7', ' ', '2', '0', '1', '8', '-',
                         '0', '3', '-', '1', '4', 'V', '0', '.', '0', '7', ' ',
                         '2', '0', '1', '8', '-', '0', '3', '-', '1', '4',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                         '\x00', '\x00', '\x00', '\x00', '\x00', '\x00']

        var = Variable(1, False, Types.T_CHAR, 128)
        self.assertEqual(var.value_to_load(value), expected_load)

    def test_arr_float_value_to_load(self):
        """Test value to load method."""
        value = [1.0, 2.0, 3.0, 4.0]
        expected_load = []
        for v in value:
            expected_load.extend(list(map(chr, struct.pack('<f', v))))
        var = Variable(1, False, Types.T_FLOAT, 4)
        self.assertEqual(var.value_to_load(value), expected_load)

    def test_int_load_to_value(self):
        """Test load to value conversion method."""
        expected_value = 257
        load = [chr(1) for _ in range(2)]
        var = Variable(1, False, Types.T_UINT16, 1)
        self.assertEqual(var.load_to_value(load), expected_value)

    def test_float_load_to_value(self):
        """Test load to value conversion method."""
        load = [chr(v) for v in struct.pack('<f', 10.5)]
        expected_value = 10.5
        var = Variable(1, False, Types.T_FLOAT, 1)
        self.assertEqual(var.load_to_value(load), expected_value)

    def test_string_load_to_value(self):
        """Test load to value conversion method."""
        load = ['V', '0', '.', '0', '7', ' ', '2', '0', '1', '8', '-', '0',
                '3', '-', '1', '4', 'V', '0', '.', '0', '7', ' ', '2', '0',
                '1', '8', '-', '0', '3', '-', '1', '4', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00', '\x00',
                '\x00', '\x00', '\x00', '\x00', '\x00']
        expected_value = [v.encode() for v in load]

        var = Variable(1, False, Types.T_CHAR, 128)
        self.assertEqual(var.load_to_value(load), expected_value)

    def test_arr_float_load_to_value(self):
        """Test load to value conversion method."""
        load = ['\x00', '\x00', '\x80', '?', '\x00', '\x00', '\x80', '?',
                '\x00', '\x00', '\x80', '?', '\x00', '\x00', '\x00', '\x00']
        expected_value = [1.0, 1.0, 1.0, 0.0]
        var = Variable(1, False, Types.T_FLOAT, 4)
        self.assertEqual(var.load_to_value(load), expected_value)


class TestBSMPVariablesGroup(TestCase):
    """Test VariablesGroup class."""

    api = (
        'ALL',
        'READ_ONLY',
        'WRITEABLE',
        'load_to_value',
        'value_to_load',
        'variables_size',
    )

    def setUp(self):
        """Common setup for all test."""
        variables = [
            Variable(0, False, Types.T_UINT8, 1),
            Variable(1, False, Types.T_UINT16, 1),
            Variable(2, False, Types.T_UINT32, 1),
            Variable(3, False, Types.T_FLOAT, 1),
            Variable(4, False, Types.T_FLOAT, 4),
            Variable(5, False, Types.T_CHAR, 8),
        ]

        self.group = VariablesGroup(3, False, variables)

    def test_api(self):
        """Test API."""
        self.assertTrue(check_public_interface_namespace(VariablesGroup, self.api))

    def _conv_value(self, fmt, value):
        return list(map(chr, struct.pack(fmt, value)))

    def test_load_to_value(self):
        """Test load conversion to value."""
        c = self._conv_value
        string = ['t', 'e', 's', 't', 'e', chr(0), chr(0), chr(0)]
        load = c('<b', 5) + c('<h', 15) + c('<i', 25) + c('<f', 1.5) + \
            c('<f', 1.6) + c('<f', 1.7) + c('<f', 1.8) + c('<f', 1.9) + \
            string
        expected_value = [5, 15, 25, 1.5, [1.6, 1.7, 1.8, 1.9],
                          [b't', b'e', b's', b't', b'e', b'\x00', b'\x00',
                          b'\x00']]
        value = self.group.load_to_value(load)
        self.assertEqual(value[:4], expected_value[:4])
        for i, f in enumerate(value[4]):
            self.assertAlmostEqual(f, expected_value[4][i])
        self.assertEqual(value[-1], expected_value[-1])

    def test_value_to_load(self):
        """Test value conversion to load."""
        c = self._conv_value
        string = ['t', 'e', 's', 't', 'e', chr(0), chr(0), chr(0)]
        expected_load = c('<b', 5) + c('<h', 15) + c('<i', 25) + \
            c('<f', 1.5) + c('<f', 1.6) + c('<f', 1.7) + c('<f', 1.8) + \
            c('<f', 1.9) + string
        value = [5, 15, 25, 1.5, [1.6, 1.7, 1.8, 1.9], 'teste']
        load = self.group.value_to_load(value)
        self.assertEqual(load, expected_load)


class TestWfmCurve(TestCase):
    """Test WfmRef Curve class."""


class TestBSMPFunction(TestCase):
    """Test Function class."""

    api = ("value_to_load", "load_to_value")

    def test_api(self):
        """Test API."""
        self.assertTrue(check_public_interface_namespace(Function, self.api))

    # def test_small_input_size(self):
    #     """Test input size smaller than 0."""
    #     with self.assertRaises(ValueError):
    #         Function(1, -1, Types.T_UINT16, 1, Types.T_UINT16)

    # def test_small_output_size(self):
    #     """Test output size smaller than 0."""
    #     with self.assertRaises(ValueError):
    #         Function(1, [Types.T_UINT16], [])

    def test_big_input_size(self):
        """Test input size bigger than 0."""
        with self.assertRaises(ValueError):
            Function(1, [Types.T_UINT16 for _ in range(65)], [Types.T_UINT16])

    def test_big_putput_size(self):
        """Test output size bigger than 0."""
        with self.assertRaises(ValueError):
            Function(1, [Types.T_UINT16], [Types.T_UINT16 for _ in range(44)])

    def test_null_value_to_load(self):
        """Test empty list is returned."""
        ret = Function(1, [Types.T_UINT16], [Types.T_UINT16]).value_to_load(None)
        self.assertEqual(ret, [])

    def test_strange_value_to_load(self):
        """Test empty list is returned."""
        with self.assertRaises(TypeError):
            Function(1, [Types.T_UINT16], [Types.T_UINT16]).value_to_load(1.5)
        with self.assertRaises(TypeError):
            Function(1, [Types.T_CHAR], [Types.T_UINT16]).value_to_load(15)

    def test_int_value_to_load(self):
        """Test value to load method."""
        value = 10
        expected_load = [chr(b) for b in struct.pack('<b', 10)]
        func = Function(1, [Types.T_UINT8], [Types.T_UINT8])
        self.assertEqual(func.value_to_load(value), expected_load)

    def test_float_value_to_load(self):
        """Test value to load method."""
        value = 10.5
        expected_load = [chr(b) for b in struct.pack('<f', value)]
        func = Function(1, [Types.T_FLOAT], [Types.T_UINT8])
        self.assertEqual(func.value_to_load(value), expected_load)

    def test_string_value_to_load(self):
        """Test value to load method."""
        value = 'V0.07 2018-03'
        expected_load = ['V', '0', '.', '0', '7', ' ', '2', '0', '1', '8', '-',
                         '0', '3', '\x00', '\x00']

        func = Function(1, [Types.T_CHAR for _ in range(15)], [Types.T_UINT8])
        self.assertEqual(func.value_to_load(value), expected_load)

    def test_arr_float_value_to_load(self):
        """Test value to load method."""
        value = [1.0, 2.0]
        expected_load = []
        for v in value:
            expected_load.extend(list(map(chr, struct.pack('<f', v))))
        func = Function(1, [Types.T_FLOAT, Types.T_FLOAT], [Types.T_UINT8])
        self.assertEqual(func.value_to_load(value), expected_load)

    def test_composite_value_to_load(self):
        """Test value to load method when there are many types involved."""
        values = [2032018, 15.41]
        load = ['\x92', '\x01', '\x1f', '\x00', '\\', '\x8f', 'v', 'A']
        func = Function(1, [Types.T_UINT32, Types.T_FLOAT], [Types.T_UINT8])
        self.assertEqual(func.value_to_load(values), load)

    def test_null_load_to_value(self):
        """Test empty list is returned."""
        ret = Function(1, [Types.T_UINT8], [Types.T_UINT8]).load_to_value(None)
        self.assertIsNone(ret)

    def test_empty_load_to_value(self):
        """Test empty list is returned."""
        ret = Function(1, [Types.T_UINT8], [Types.T_UINT8]).load_to_value([])
        self.assertIsNone(ret)

    def test_int_load_to_value(self):
        """Test load to value conversion method."""
        expected_value = 1
        load = [chr(1)]
        func = Function(1, [Types.T_UINT8], [Types.T_UINT8])
        self.assertEqual(func.load_to_value(load), expected_value)

    def test_float_load_to_value(self):
        """Test load to value conversion method."""
        load = [chr(v) for v in struct.pack('<f', 10.5)]
        expected_value = 10.5
        func = Function(1, [Types.T_UINT8], [Types.T_FLOAT])
        self.assertEqual(func.load_to_value(load), expected_value)

    def test_string_load_to_value(self):
        """Test load to value conversion method."""
        load = ['V', '0', '.', '0', '7', ' ', '2', '0', '1', '8', '\x00',
                '\x00', '\x00', '\x00', '\x00']

        expected_value = [b'V', b'0', b'.', b'0', b'7', b' ', b'2', b'0', b'1',
                          b'8', b'\x00', b'\x00', b'\x00', b'\x00', b'\x00']
        func = Function(1, [Types.T_UINT8], [Types.T_CHAR for _ in range(15)])
        self.assertEqual(func.load_to_value(load), expected_value)

    def test_arr_float_load_to_value(self):
        """Test load to value conversion method."""
        load = ['\x00', '\x00', '\x80', '?', '\x00', '\x00', '\x00', '\x00']
        expected_value = [1.0, 0.0]
        func = Function(1, [Types.T_UINT8], [Types.T_FLOAT, Types.T_FLOAT])
        self.assertEqual(func.load_to_value(load), expected_value)

    def test_composite_load_to_value(self):
        """Test load to value when there are many types involved."""
        values = [2032018, 15.409999847]
        load = ['\x92', '\x01', '\x1f', '\x00', '\\', '\x8f', 'v', 'A']
        func = Function(1, [Types.T_UINT8], [Types.T_UINT32, Types.T_FLOAT])
        self.assertEqual(func.load_to_value(load)[0], values[0])
        self.assertAlmostEqual(func.load_to_value(load)[1], values[1])


class TestBSMPEntities(TestCase):
    """This class only has propertyies."""



"""BeagleBone class."""

# NOTE on current behaviour of BeagleBone:
#
# 01. While in RmpWfm, MigWfm or SlowRefSync, the PS_I_LOAD variable read from
#     power supplies after setting the last curve point may not be the
#     final value given by PS_REFERENCE. This is due to the fact that the
#     power supply control loop takes some time to converge and the PRU may
#     block serial comm. before it. This is evident in SlowRefSync mode, where
#     reference values may change considerably between two setpoints.
#     (see identical note in PRUController)
#
# 02. Now that SOFB controls correctors power supplies directly, the
#     'priority_pvs' mechanism for updating SOFBCurrent-* PVs may be dropped.

import time as _time

from ..devices import StrengthConv as _StrengthConv


class BeagleBone:
    """BeagleBone class.

    Through objects of this class power supply controllers can be accessed
    in order to redirect read, write connected requests to the
    aproppriate device.
    """

    def __init__(self, controllers, databases):
        """Init object.

        Parameters:
        controllers : list of PSController objectsused to communicate with
                      BSMP devices.
        databases   : corresponding controllers databases.

        """
        # ps controllers.
        self._controllers = controllers

        # ps cpntrollers databases
        self._databases = databases

        # psnames
        self._psnames = tuple(self._controllers.keys())

        # strength property names
        self._strenames = self._get_strength_names()

        # create devname to scan interval dict
        self._create_dev2interval_dict()

        # init mirror variables and last update timestamp dicts
        self._create_dev2mirr_dev2timestamp_dict()

        # create strength conv epics objects
        self._streconvs, self._streconnected, self._strelims = \
            self._create_streconvs()

        # initialized state
        self._initialized = False

    @property
    def controllers(self):
        """Return beaglebone power supply controllers."""
        return self._controllers

    @property
    def psnames(self):
        """PS names."""
        return self._psnames

    def update_interval(self, devname=None):
        """Update interval, as defined in PRUcontrollers."""
        if devname is not None:
            return self._dev2interval[devname]
        intervals = tuple(self._dev2interval.values())
        return min(intervals)

    def read(self, devname, field=None, force_update=False):
        """Read from device."""
        # if not initialized, return None
        if not self._initialized:
            return None, False

        now = _time.time()
        last = self._dev2timestamp[devname]

        # NOTE: update frequency with which class updates state mirror of
        # power supply.
        interval = self._dev2interval[devname]

        # reads, if updating is needed
        if force_update or last is None or now - last > interval:
            updated = True
            self._dev2mirror[devname] = \
                self._controllers[devname].read_all_fields(devname)
            self._update_strengths(devname)
            self._dev2timestamp[devname] = now
        else:
            updated = False

        if field is None:
            return self._dev2mirror[devname], updated
        else:
            pvname = devname + ':' + field
            return self._dev2mirror[devname][pvname], updated

    def write(self, devname, field, value):
        """Write value to a device field.

        Return pvname-value dictionary of priority PVs.
        """
        # if not initialized, return None
        if not self._initialized:
            return None

        if field in {'Energy-SP', 'Kick-SP', 'KL-SP', 'SL-SP'}:
            streconv = self._streconvs[devname]
            curr = streconv.conv_strength_2_current(value)
            priority_pvs = self._controllers[devname].write(
                devname, 'Current-SP', curr)
        else:
            priority_pvs = self._controllers[devname].write(
                devname, field, value)
        return priority_pvs

    def get_strength_limits(self, devname):
        """Return strength lower and upper limits."""
        return self._strelims[devname]

    def check_connected(self, devname):
        """Check wether device is connected."""
        return self._controllers[devname].check_connected(devname)

    def check_connected_strength(self, devname):
        """Check connection with PVs for strength calc."""
        return self._streconnected[devname]

    def strength_name(self, devname):
        """Return strength name."""
        return self._strenames[devname]

    def strength_limits(self, devname):
        """Return strength limits."""
        return self._strelims[devname]

    def database(self, devname):
        """Return device database."""
        return self._databases[devname]

    def init(self, processing=True, scanning=True):
        """Initialize controllers."""
        # return  # allow for IOC initialization without HW comm.

        # initialize controller communication and setpoint fields
        pruc_initialized = set()
        psc_initialized = set()
        for controller in self._controllers.values():
            pruc = controller.pru_controller
            if pruc not in pruc_initialized:
                pruc.bsmp_init_communication()
                pruc.processing = processing
                pruc.scanning = scanning
                pruc_initialized.add(pruc)
            if controller not in psc_initialized:
                controller.init_setpoints()
                psc_initialized.add(controller)

        # guarantee strengthconv PVs connected
        for streconv in self._streconvs.values():
            streconv.wait_for_connection()

        self._initialized = True

    # --- private methods ---

    def _get_strength_names(self):
        strenames = dict()
        for psname, dbase in self._databases.items():
            if 'Energy-SP' in dbase:
                strenames[psname] = 'Energy'
            elif 'Kick-SP' in dbase:
                strenames[psname] = 'Kick'
            elif 'KL-SP' in dbase:
                strenames[psname] = 'KL'
            elif 'SL-SP' in dbase:
                strenames[psname] = 'SL'
            else:
                strenames[psname] = None
        return strenames

    def _create_dev2mirr_dev2timestamp_dict(self):
        self._dev2timestamp = dict()
        self._dev2mirror = dict()
        for devname in self._controllers:
            self._dev2timestamp[devname] = None
            self._dev2mirror[devname] = dict()

    def _create_dev2interval_dict(self):
        self._dev2interval = dict()
        for devname, controller in self._controllers.items():
            pruc = controller.pru_controller
            self._dev2interval[devname] = 1.0/pruc.params.FREQ_SCAN

    def _create_streconvs(self):
        streconvs = dict()
        strec = dict()
        strelims = dict()
        for psname in self.psnames:
            if 'DCLink' in psname or psname.startswith('IT'):
                strec[psname] = True
            else:
                # NOTE: use 'Ref-Mon' proptype for all
                strec[psname] = False
                streconvs[psname] = _StrengthConv(psname, 'Ref-Mon')
                strelims[psname] = [None, None]
        return streconvs, strec, strelims

    def _update_strengths(self, psname):
        # t0_ = _time.time()
        if 'DCLink' in psname or psname.startswith('IT'):
            return
        streconv = self._streconvs[psname]
        strelims = self._strelims[psname]
        mirror = self._dev2mirror[psname]
        dbase = self._databases[psname]
        curr0 = mirror[psname + ':Current-SP']
        curr1 = mirror[psname + ':Current-RB']
        curr2 = mirror[psname + ':CurrentRef-Mon']
        curr3 = mirror[psname + ':Current-Mon']
        curr4 = dbase['Current-SP']['lolo']
        curr5 = dbase['Current-SP']['hihi']
        currs = (curr0, curr1, curr2, curr3, curr4, curr5)
        strengths = streconv.conv_current_2_strength(currents=currs)
        if strengths is None or None in strengths:
            self._streconnected[psname] = False
        else:
            self._streconnected[psname] = True
            propname = psname + ':' + self._strenames[psname]
            mirror[propname + '-SP'] = strengths[0]
            mirror[propname + '-RB'] = strengths[1]
            mirror[propname + 'Ref-Mon'] = strengths[2]
            mirror[propname + '-Mon'] = strengths[3]
            # update strength limits
            if strengths[4] <= strengths[5]:
                strelims[0], strelims[1] = strengths[4], strengths[5]
            else:
                strelims[0], strelims[1] = strengths[5], strengths[4]
        # t1_ = _time.time()
        # print('update_strengths: {:.3f}'.format(1000*(t1_-t0_)))



"""Ramp utility module."""

# nominal strength values
BO_INJECTION_ENERGY = 0.150  # [GeV]
BO_EJECTION_ENERGY = 3.0  # [GeV]

BO_INJECTION_RF_GAPVOLT = 150.0074530757551  # [kV]
BO_EJECTION_RF_GAPVOLT = 950.0  # [kV]

NOMINAL_STRENGTHS = {
    'SI-Fam:PS-B1B2-1': BO_EJECTION_ENERGY,  # [Energy: GeV]
    'SI-Fam:PS-B1B2-2': BO_EJECTION_ENERGY,  # [Energy: GeV]
    'SI-Fam:PS-QFA': +0.7146305692912001,  # [KL: 1/m]
    'SI-Fam:PS-QDA': -0.2270152048045000,  # [KL: 1/m]
    'SI-Fam:PS-QFB': +1.2344424683922000,  # [KL: 1/m]
    'SI-Fam:PS-QDB2': -0.4782973132726601,  # [KL: 1/m]
    'SI-Fam:PS-QDB1': -0.2808906119138000,  # [KL: 1/m]
    'SI-Fam:PS-QFP': +1.2344424683922000,  # [KL: 1/m]
    'SI-Fam:PS-QDP2': -0.4782973132726601,  # [KL: 1/m]
    'SI-Fam:PS-QDP1': -0.2808906119138000,  # [KL: 1/m]
    'SI-Fam:PS-Q1': +0.5631612043340000,  # [KL: 1/m]
    'SI-Fam:PS-Q2': +0.8684629376249999,  # [KL: 1/m]
    'SI-Fam:PS-Q3': +0.6471254242426001,  # [KL: 1/m]
    'SI-Fam:PS-Q4': +0.7867827142062001,  # [KL: 1/m]
    'SI-Fam:PS-SDA0': -12.1250549999999979,  # [SL: 1/m^2]
    'SI-Fam:PS-SDB0': -09.7413299999999996,  # [SL: 1/m^2]
    'SI-Fam:PS-SDP0': -09.7413299999999996,  # [SL: 1/m^2]
    'SI-Fam:PS-SDA1': -24.4479749999999996,  # [SL: 1/m^2]
    'SI-Fam:PS-SDB1': -21.2453849999999989,  # [SL: 1/m^2]
    'SI-Fam:PS-SDP1': -21.3459000000000003,  # [SL: 1/m^2]
    'SI-Fam:PS-SDA2': -13.3280999999999992,  # [SL: 1/m^2]
    'SI-Fam:PS-SDB2': -18.3342150000000004,  # [SL: 1/m^2]
    'SI-Fam:PS-SDP2': -18.3421500000000002,  # [SL: 1/m^2]
    'SI-Fam:PS-SDA3': -20.9911199999999987,  # [SL: 1/m^2]
    'SI-Fam:PS-SDB3': -26.0718599999999974,  # [SL: 1/m^2]
    'SI-Fam:PS-SDP3': -26.1236099999999993,  # [SL: 1/m^2]
    'SI-Fam:PS-SFA0': +7.8854400000000000,  # [SL: 1/m^2]
    'SI-Fam:PS-SFB0': +11.0610149999999994,  # [SL: 1/m^2]
    'SI-Fam:PS-SFP0': +11.0610149999999994,  # [SL: 1/m^2]
    'SI-Fam:PS-SFA1': +28.7742599999999982,  # [SL: 1/m^2]
    'SI-Fam:PS-SFB1': +34.1821950000000001,  # [SL: 1/m^2]
    'SI-Fam:PS-SFP1': +34.3873949999999979,  # [SL: 1/m^2]
    'SI-Fam:PS-SFA2': +22.6153800000000018,  # [SL: 1/m^2]
    'SI-Fam:PS-SFB2': +29.6730900000000020,  # [SL: 1/m^2]
    'SI-Fam:PS-SFP2': +29.7755099999999970,  # [SL: 1/m^2]
    'BO-Fam:PS-B-1': BO_INJECTION_ENERGY,  # [Energy: GeV]
    'BO-Fam:PS-B-2': BO_INJECTION_ENERGY,  # [Energy: GeV]
    'BO-Fam:PS-QD': +0.0011197961538728,  # [KL: 1/m]
    'BO-Fam:PS-QF': +0.3770999232791374,  # [KL: 1/m]
    'BO-Fam:PS-SD': +0.5258382119529604,  # [SL: 1/m^2]
    'BO-Fam:PS-SF': +1.1898514030258744,  # [SL: 1/m^2]
}

# Power supply ramp parameters
DEFAULT_PS_RAMP_DURATION = 490.0  # [ms]

DEFAULT_PS_RAMP_RAMPUP1_START_TIME = 10.0  # [ms]
DEFAULT_PS_RAMP_RAMPUP2_START_TIME = 140.0  # [ms]
DEFAULT_PS_RAMP_RAMPDOWN_START_TIME = 378.0  # [ms]
DEFAULT_PS_RAMP_RAMPDOWN_STOP_TIME = 488.0  # [ms]
DEFAULT_PS_RAMP_RAMPUP_SMOOTH_INTVL = 60.0  # [ms]
DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_INTVL = 60.0  # [ms]

# --- Current values [A] ---
DEFAULT_PS_RAMP_START_CURRENT = 35.0
DEFAULT_PS_RAMP_RAMPUP1_START_CURRENT = 42.0
DEFAULT_PS_RAMP_RAMPUP2_START_CURRENT = 199.7
DEFAULT_PS_RAMP_RAMPDOWN_START_CURRENT = 1140.7
DEFAULT_PS_RAMP_RAMPDOWN_STOP_CURRENT = 42.0
DEFAULT_PS_RAMP_RAMPUP_SMOOTH_CURRENT = 5.7
DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_CURRENT = 91.3

# --- Energy values [GeV] ---
# NOTE: these constants are derived from excitation data
# current 150mev = 50.625127792833354
# current 3gev = 1013.2431132714938
DEFAULT_PS_RAMP_START_ENERGY = 0.10334265034313583
DEFAULT_PS_RAMP_RAMPUP1_START_ENERGY = 0.12409653823817533
DEFAULT_PS_RAMP_RAMPUP2_START_ENERGY = 0.5994074098589306
DEFAULT_PS_RAMP_RAMPDOWN_START_ENERGY = 3.328600085387852
DEFAULT_PS_RAMP_RAMPDOWN_STOP_ENERGY = 0.12409653823817533
DEFAULT_PS_RAMP_RAMPUP_SMOOTH_ENERGY = 0.01683008877016784
DEFAULT_PS_RAMP_RAMPDOWN_SMOOTH_ENERGY = 0.27256207120794684


# Timing parameters
DEFAULT_TI_PARAMS_INJECTION_TIME = 17.11  # [ms]
DEFAULT_TI_PARAMS_EJECTION_TIME = 345.76  # [ms]

DEFAULT_TI_PARAMS_PS_RAMP_DELAY = 0.0  # [us]
DEFAULT_TI_PARAMS_RF_RAMP_DELAY = DEFAULT_PS_RAMP_RAMPUP1_START_TIME  # [us]


# RF ramp parameters
MAX_RF_RAMP_DURATION = 500.0  # [ms]
DEFAULT_RF_RAMP_BOTTOM_DURATION = 0.0  # [ms]
DEFAULT_RF_RAMP_RAMPUP_DURATION = 340.0  # [ms]
DEFAULT_RF_RAMP_TOP_DURATION = 10.0  # [ms]
DEFAULT_RF_RAMP_RAMPDOWN_DURATION = 55.0  # [ms]
_DURATION = DEFAULT_RF_RAMP_BOTTOM_DURATION + \
            DEFAULT_RF_RAMP_RAMPUP_DURATION + \
            DEFAULT_RF_RAMP_TOP_DURATION + \
            DEFAULT_RF_RAMP_RAMPDOWN_DURATION
if _DURATION > MAX_RF_RAMP_DURATION:
    raise ValueError('Invalid RF ramp default durations.')
del _DURATION


def _calc_bot_top_voltage():
    # # Linear extrapolation to start and stop rf gap voltages
    _vt1 = DEFAULT_TI_PARAMS_INJECTION_TIME
    _vt2 = DEFAULT_TI_PARAMS_EJECTION_TIME
    _vv1 = BO_INJECTION_RF_GAPVOLT
    _vv2 = BO_EJECTION_RF_GAPVOLT
    _vmm = (_vv2 - _vv1)/(_vt2 - _vt1)
    _vtt = DEFAULT_TI_PARAMS_RF_RAMP_DELAY + DEFAULT_RF_RAMP_BOTTOM_DURATION
    def_bottom = _vv1 + _vmm*(_vtt - _vt1)  # [kV]
    _vtt = _vtt + DEFAULT_RF_RAMP_RAMPUP_DURATION
    def_top = _vv1 + _vmm*(_vtt - _vt1)  # [kV]
    return def_bottom, def_top


DEFAULT_RF_RAMP_BOTTOM_VOLTAGE, DEFAULT_RF_RAMP_TOP_VOLTAGE = \
    _calc_bot_top_voltage()

DEFAULT_RF_RAMP_BOTTOM_PHASE = 0.0  # [°]
DEFAULT_RF_RAMP_TOP_PHASE = 0.0  # [°]


def update_nominal_strengths(dic):
    """Update dictionary with nominal values."""
    for k, value in NOMINAL_STRENGTHS.items():
        if k in dic:
            dic[k] = value



"""Base Simulator."""

import random as _random
from abc import ABC, abstractmethod
import numpy as _np

from . import DBASE_DEFAULT as _DBASE_DEF


class Simulator(ABC):
    """Simulator.

    Independent and markovian PVs simulator,

    dbase_default : epics database used as default if no other dbase
                    for pvname regular expression is initialized.

    """

    class Utils:
        """Utils class."""

        @staticmethod
        def add_fluctuations(value, relative=0, absolute=0):
            """."""
            if isinstance(value, _np.ndarray):
                rnd_rel = _np.random.uniform(
                    1-relative, 1+relative, len(value)) if relative else 1
                rnd_abs = _np.random.uniform(
                    -absolute, +absolute, len(value)) if absolute else 0
            else:
                rnd_rel = _random.uniform(
                    1-relative, 1+relative) if relative else 1
                rnd_abs = _random.uniform(
                    -absolute, +absolute) if absolute else 0

            value_new = value * rnd_rel + rnd_abs
            return value_new

    _DBASE_DEFAULT = _DBASE_DEF

    def __init__(self, dbase_default=None):
        """."""
        self._dbase_default = \
            dbase_default if dbase_default else Simulator._DBASE_DEFAULT

        # simulator stores it own dict of
        # used SimPVs for optimization
        self._pvs = dict()

    # --- abstract classes that subclasses MUST implement ---

    @abstractmethod
    def callback_pv_dbase(self):
        """Return dict of pvname regular expression and database.

        This method is called when simulator is registered in simulation.
        """

    @abstractmethod
    def callback_pv_add(self, pvname):
        """Execute callback method after a new SimPV is added to simulator.

        This method is called when a new SimPV is created and it is used in
        the simulator, as implicit defined in dbase initializations.
        """

    @abstractmethod
    def callback_pv_get(self, pvname, **kwargs):
        """Execute callback function prior to SimPV readout.

        This method is called when a SimPV value is requested.
        """

    @abstractmethod
    def callback_pv_put(self, pvname, value, **kwargs):
        """Execute callback function prior to SimPV setpoint.

        Return True if setpoint is acceptable, False otherwise.
        """

    @abstractmethod
    def callback_update(self, **kwargs):
        """Execute callback to update/synchronize simulator.

        This method is called when simulation state update is required.
        """

    # --- general methods ---

    @property
    def pvnames(self):
        """Return name of SimPVs of simulator."""
        return set(self._pvs.keys())

    @property
    def values(self):
        """Return dict with pvnames and associated values of simulator."""
        vals = dict()
        for pvname in self._pvs:
            vals[pvname] = self.pv_value_get(pvname)
        return vals

    def pv_check(self, pvname):
        """Check if SimPV belongs to simulator."""
        dbases = self.callback_pv_dbase()
        for regexp in dbases:
            if regexp.match(pvname):
                return True
        return False

    def pv_value_get(self, pvname):
        """Get SimPV value without invoking simulator callback."""
        return self._pvs[pvname].get_sim()

    def pv_value_put(self, pvname, value):
        """Set SimPV value without invoking simulator callback."""
        self._pvs[pvname].put_sim(value)

    def update(self, **kwargs):
        """Update simulator."""
        self.callback_update(**kwargs)

    def reset(self):
        """Reset simulator."""
        self._pvs = dict()

    def __contains__(self, pvname):
        """Return True if SimPV is in simulator."""
        return pvname in self._pvs

    # --- private methods ---

    # NOTE: not a _-private method, but its use should be,
    # accessable only by friend class Simulation.
    def callback_pv_register(self, pvobj):
        """Register SimPB.

        This method is be used solenely by Simulation class methods.
        """
        pvname = pvobj.pvname
        self._pvs[pvname] = pvobj
        self.callback_pv_add(pvname)



"""ID Search module."""

from threading import Lock as _Lock

from .. import clientweb as _web
from ..magnet.excdata import ExcitationData as _ExcitationData

from .ps_search import PSSearch as _PSSearch


class IDSearch:
    """ID Search Class."""

    # NOTE: Maybe we should move these mappings to
    # static tables in our cs-constants server.

    # NOTE: see
    # https://wiki-sirius.lnls.br/mediawiki/index.php/Machine:Insertion_Devices

    _beamline2idname = {
        'CARNAUBA': 'SI-06SB:ID-APU22',
        'CATERETE': 'SI-07SP:ID-APU22',
        'EMA':      'SI-08SB:ID-APU22',
        'MANACA':   'SI-09SA:ID-APU22',
        'SABIA':    'SI-10SB:ID-Delta52',
        'IPE':      'SI-11SP:ID-APU58',
        # 'PAINEIRA': 'SI-14SB:ID-???',
        # 'SAPUCAIA': 'SI-17SA:ID-???',
    }
    _idname2beamline = {v: k for k, v in _beamline2idname.items()}

    _idname_2_orbitcorr = {}

    _idname_2_orbitffwd_fname = {}

    _idname_2_orbitffwd_dict = dict()

    _lock = _Lock()

    @staticmethod
    def conv_idname_2_beamline(idname):
        """Return the beam line name corresponding to sector."""
        if idname in IDSearch._idname2beamline.keys():
            return IDSearch._idname2beamline[idname]
        return None

    @staticmethod
    def conv_beamline_2_idname(beamline):
        """Return the beam line name corresponding to idname."""
        if beamline in IDSearch._beamline2idname.keys():
            return IDSearch._beamline2idname[beamline]
        return None

    @staticmethod
    def get_idname_2_beamline_dict():
        """Return idname to beamline dictionary."""
        return IDSearch._idname2beamline.copy()

    @staticmethod
    def get_beamline_2_idname_dict():
        """Return beamline to idname dictionary."""
        return IDSearch._beamline2idname.copy()

    @staticmethod
    def get_idnames(filters=None):
        """Return a sorted and filtered list of all ID names."""
        filters_ = dict()
        if filters:
            filters_.update(filters)
        filters_['dis'] = 'ID'
        return _PSSearch.get_psnames(filters_)

    @staticmethod
    def conv_idname_2_orbitcorr(idname):
        """Return list of orbit correctors of a given ID."""
        return IDSearch._idname_2_orbitcorr[idname]

    @staticmethod
    def conv_idname_2_orbitffwd(idname):
        """Convert idname to orbit feedforward excdata."""
        IDSearch._reload_idname_2_orbitffwd_dict(idname)
        return IDSearch._idname_2_orbitffwd_dict[idname]

    # --- private methods ---

    @staticmethod
    def _reload_idname_2_orbitffwd_dict(idname):
        """Load ID ffwd data."""
        with IDSearch._lock:
            if idname in IDSearch._idname_2_orbitffwd_dict:
                return
            if not _web.server_online():
                raise Exception(
                    'could not read "' + str(idname) + '" from web server!')
            ffwd_fname = IDSearch._idname_2_orbitffwd_fname[idname]
            IDSearch._idname_2_orbitffwd_dict[idname] = \
                _ExcitationData(filename_web=ffwd_fname + '.txt')
